import logging
import math as m
from typing import Optional, Sequence, Tuple, Union, cast

import numpy as np
import pandas as pd
import sqlalchemy as sqla
import torch as th
import torch.nn.functional as F
import zarr
from numpy.typing import ArrayLike
from skimage.measure import regionprops_table
from sqlalchemy.orm import Session
from tqdm import tqdm
from zarr.storage import StoreLike

from ultrack.config import MainConfig
from ultrack.core.database import NodeDB
from ultrack.utils.array import create_zarr
from ultrack.utils.constants import ULTRACK_DEBUG
from ultrack.utils.cuda import import_module, torch_default_device, xp

logging.basicConfig()
logging.getLogger("sqlachemy.engine").setLevel(logging.INFO)

LOG = logging.getLogger(__name__)


_ALIGN_CORNERS = True


def _interpolate(tensor: th.Tensor, antialias: bool = False, **kwargs) -> th.Tensor:
    """Interpolates tensor.

    Parameters
    ----------
    tensor : th.Tensor
        Input 4 or 5-dim tensor.
    antialias : bool, optional
        When true applies a gaussian filter with 0.5 * downscale factor.
        Ignored if upscaling.

    Returns
    -------
    th.Tensor
        Interpolated tensor.
    """
    mode = "trilinear" if tensor.ndim == 5 else "bilinear"
    if antialias:
        scale_factor = kwargs.get("scale_factor")
        if scale_factor is None:
            raise ValueError(
                "`_interpolate` with `antialias=True` requires `scale_factor` parameter."
            )

        if scale_factor < 1.0:
            ndi = import_module("scipy", "ndimage")
            orig_shape = tensor.shape

            if xp == np:  # xp ideally should be cupy in this case
                LOG.warning(
                    "Cupy not found, using scipy for interpolation. This may slow down ultrack."
                )
                array = xp.asarray(tensor.cpu().squeeze().contiguous().numpy())
            else:
                array = xp.asarray(tensor.squeeze().contiguous())
            blurred = ndi.gaussian_filter(
                array,
                sigma=0.5 / scale_factor,
                output=array,
            )
            tensor = th.as_tensor(blurred, device=tensor.device)
            tensor = tensor.reshape(orig_shape)
            LOG.info(f"Antialiasing with sigma = {0.5 / scale_factor}.")

    return F.interpolate(tensor, **kwargs, mode=mode, align_corners=_ALIGN_CORNERS)


def total_variation_loss(tensor: th.Tensor) -> th.Tensor:
    loss = 0.0
    for i in range(1, tensor.ndim - 1):
        idx = th.arange(tensor.shape[i], device=tensor.device)
        # tv = th.square(
        #     2 * th.index_select(tensor, i, idx[1:-1]) \
        #       - th.index_select(tensor, i, idx[2:]) \
        #       - th.index_select(tensor, i, idx[:-2])
        # )  # second derivative
        tv = th.square(
            th.index_select(tensor, i, idx[:-1]) - th.index_select(tensor, i, idx[1:])
        )  # first derivative

        loss = loss + tv.sum()
    return loss


def identity_grid(shape: tuple[int, ...]) -> th.Tensor:
    """Grid equivalent to a identity vector field (no flow).

    Parameters
    ----------
    shape : tuple[int, ...]
        Grid shape.

    Returns
    -------
    th.Tensor
        Tensor of shape (Z, Y, X, D)
    """
    ndim = len(shape)
    T = th.zeros((1, ndim, ndim + 1))
    T[:, :, :-1] = th.eye(ndim)
    grid_shape = (1, 1) + shape
    return F.affine_grid(T, grid_shape, align_corners=_ALIGN_CORNERS)


def _interpolate_grid(grid: th.Tensor, out_dim: int = -1, **kwargs) -> th.Tensor:
    """Interpolate a grid.

    Parameters
    ----------
    grid : th.Tensor
        Grid tensor of shape (z, y, x, D).
    out_dim : int, optional
        Output dimension, by default -1.

    Returns
    -------
    th.Tensor
        Tensor of shape (Z, Y, X, D) when out_dim=1, D position is subject to `out_dim` value.
    """
    return th.stack(
        [_interpolate(grid[None, ..., d], **kwargs)[0] for d in range(grid.shape[-1])],
        dim=out_dim,
    )


def flow_field(
    source: th.Tensor,
    target: th.Tensor,
    im_factor: int = 4,
    grid_factor: int = 4,
    num_iterations: int = 1000,
    lr: float = 1e-2,
    n_scales: int = 3,
    init_grid: Optional[th.Tensor] = None,
) -> th.Tensor:
    """
    Compute the flow vector field `T` that minimizes the
    mean squared error between `T(source)` and `target`.

    Parameters
    ----------
    source : torch.Tensor
        Source image (C, Z, Y, X).
    target : torch.Tensor
        Target image (C, Z, Y, X).
    im_factor : int, optional
        Image space down scaling factor, by default 4.
    grid_factor : int, optional
        Grid space down scaling factor, by default 4.
        Grid dimensions will be divided by both `im_factor` and `grid_factor`.
    num_iterations : int, optional
        Number of gradient descent iterations, by default 1000.
    lr : float, optional
        Learning rate (gradient descent step), by default 1e-2
    n_scales : int, optional
        Number of scales used for multi-scale optimization, by default 3.

    Returns
    -------
    torch.Tensor
        Vector field array with shape (D, (Z / factor), Y / factor, X / factor)
    """
    assert source.shape == target.shape
    assert n_scales > 0
    ndim = source.ndim - 1

    if init_grid is not None:
        if init_grid.ndim != ndim + 2:
            raise ValueError(
                f"Grid must have ndim dimensions ({ndim + 2}), (B, (Z), Y, X, D). Found {init_grid.ndim}."
            )

        if init_grid.shape[-1] != ndim:
            raise ValueError(
                f"Grid last channel must match input dimensions. Found {init_grid.shape[-1]} expected {ndim}."
            )

    source = source.unsqueeze(0)
    target = target.unsqueeze(0)

    device = source.device
    scales = np.flip(np.power(2, np.arange(n_scales)))
    grid = init_grid
    norm_factor = None
    constant = 1_000  # scale of input affects the results, 1_000 is a good value

    for scale in scales:
        with th.no_grad():
            scaled_im_factor = im_factor * scale
            scaled_source = _interpolate(
                source, scale_factor=1 / scaled_im_factor, antialias=True
            )
            scaled_target = _interpolate(
                target, scale_factor=1 / scaled_im_factor, antialias=True
            )
            if norm_factor is None:
                norm_factor = (
                    2.0
                    * constant
                    / (scaled_source.quantile(0.9999) + scaled_target.quantile(0.9999))
                )

            scaled_source *= norm_factor
            scaled_target *= norm_factor

            grid_shape = tuple(
                m.ceil(s / grid_factor) for s in scaled_source.shape[-ndim:]
            )

            if grid is None:
                grid = identity_grid(grid_shape).to(device)
                grid0 = identity_grid(grid_shape).to(device)
            else:
                grid = _interpolate_grid(grid, size=grid_shape)
                grid0 = identity_grid(grid_shape).to(device)

        grid.requires_grad_(True).retain_grad()

        LOG.info(f"scale: {scale}")
        LOG.info(f"image shape: {scaled_source.shape}")
        LOG.info(f"grid shape: {grid.shape}")

        for i in range(num_iterations):
            if grid.grad is not None:
                grid.grad.zero_()

            large_grid = _interpolate_grid(grid, size=scaled_source.shape[-ndim:])

            im2hat = F.grid_sample(
                scaled_target, large_grid, align_corners=_ALIGN_CORNERS
            )

            loss = F.l1_loss(im2hat, scaled_source)
            loss = loss + total_variation_loss(grid - grid0)
            loss.backward()

            if i % 10 == 0:
                LOG.info(f"iter. {i} MSE: {loss:0.4f}")

            with th.no_grad():
                grid -= lr * grid.grad

    LOG.info(f"image size: {source.shape}")
    LOG.info(f"image factor: {im_factor}")
    LOG.info(f"grid factor: {grid_factor}")

    with th.no_grad():
        grid = cast(th.Tensor, grid)
        grid = grid - grid0
        grid = th.flip(grid, (-1,))  # x, y, z -> z, y, x

        # divided by 2.0 because the range is -1 to 1 (length = 2.0)
        grid /= 2.0
        grid = _interpolate_grid(grid, out_dim=1, size=scaled_source.shape[-ndim:])[0]

        LOG.info(f"vector field shape: {grid.shape}")

    if ULTRACK_DEBUG:
        import napari

        viewer = napari.Viewer()
        viewer.add_image(
            scaled_source.cpu().numpy(),
            name="im1",
            blending="additive",
            colormap="blue",
            visible=False,
        )
        viewer.add_image(
            scaled_target.cpu().numpy(),
            name="im2",
            blending="additive",
            colormap="green",
        )
        viewer.add_image(
            im2hat.detach().cpu().numpy(),
            name="im2hat",
            blending="additive",
            colormap="red",
        )
        viewer.add_image(
            grid.detach().cpu().numpy(),
            name="grid",
            colormap="turbo",
            visible=False,
        )
        napari.run()

    return grid


@th.no_grad()
def apply_field(field: th.Tensor, image: th.Tensor) -> th.Tensor:
    """
    Transform image using vector field.
    Image will be scaled to the field size.

    Parameters
    ----------
    field : th.Tensor
        Vector field (D, z, y, x)
    image : th.Tensor
        Original image used to compute the vector field.

    Returns
    -------
    th.Tensor
        Transformed image (z, y, x)
    """
    assert image.ndim == 4

    field = th.flip(field, (0,))  # z, y, x -> x, y, z
    field = field.movedim(0, -1)[None]

    field = field * 2.0  # mapping range from image shape to -1 to 1
    field = identity_grid(field.shape[1:-1]).to(field.device) - field

    transformed_image = F.grid_sample(image[None], field, align_corners=_ALIGN_CORNERS)

    return transformed_image[0]


@th.no_grad()
def advenct_field(
    field: ArrayLike,
    sources: th.Tensor,
    shape: Optional[tuple[int, ...]] = None,
    invert: bool = True,
) -> th.Tensor:
    """
    Advenct points from sources through the provided field.
    Shape indicates the original shape (space) and sources.
    Useful when field is down scaled from the original space.

    Parameters
    ----------
    field : ArrayLike
        Field array with shape T x D x (Z) x Y x X
    sources : th.Tensor
        Array of sources N x D
    shape : tuple[int, ...]
        When provided scales field accordingly, D-dimensional tuple.
    invert : bool
        When true flow is multiplied by -1, resulting in reversal of the flow.

    Returns
    -------
    th.Tensor
        Trajectories of sources N x T x D
    """
    ndim = field.ndim - 2
    device = sources.device
    orig_shape = th.tensor(shape, device=device)
    field_shape = th.tensor(field.shape[2:], device=device)

    if orig_shape is None:
        scales = th.ones(ndim, device=device)
    else:
        scales = (field_shape - 1) / (orig_shape - 1)

    trajectories = [sources]

    zero = th.zeros(1, device=device)

    # ignore last (first) frame since it will be empty
    iterator = range(1, field.shape[0]) if invert else range(0, field_shape[0] - 1)

    for t in iterator:
        current = th.as_tensor(field[t]).to(device=device, non_blocking=True)

        int_sources = th.round(trajectories[-1] * scales)
        int_sources = th.maximum(int_sources, zero)
        int_sources = th.minimum(int_sources, field_shape - 1).long()
        spatial_idx = tuple(
            t.T[0] for t in th.tensor_split(int_sources, len(orig_shape), dim=1)
        )
        idx = (slice(None), *spatial_idx)

        movement = current[idx].T * orig_shape

        if invert:
            sources = sources - movement
        else:
            sources = sources + movement

        trajectories.append(sources)

    trajectories = th.stack(trajectories, dim=1)

    return trajectories


def advenct_field_from_labels(
    field: ArrayLike,
    label: ArrayLike,
    invert: bool = True,
) -> ArrayLike:
    """
    Advenct points from segmentation labels centroid.

    Parameters
    ----------
    field : ArrayLike
        Field array with shape T x D x (Z) x Y x X
    label : ArrayLike
        Label image.
    invert : bool
        When true flow is multiplied by -1, resulting in reversal of the flow.

    Returns
    -------
    ArrayLike
        Trajectories of sources N x T x D
    """
    if isinstance(label, th.Tensor):
        device = label.device
        label = label.numpy()
    else:
        device = torch_default_device()

    centroids = np.stack(
        [v for v in regionprops_table(label, properties=("centroid",)).values()], axis=1
    )
    centroids = th.as_tensor(centroids, device=device)

    trajectories = advenct_field(field, centroids, shape=label.shape, invert=invert)

    trajectories = th.minimum(trajectories, th.as_tensor(label.shape, device=device))
    trajectories = th.maximum(trajectories, th.zeros(len(label.shape), device=device))

    return trajectories


def advenct_from_quasi_random(
    field: ArrayLike,
    img_shape: Tuple[int, ...],
    n_samples: int,
    invert: bool = True,
    device: Optional[th.device] = None,
) -> ArrayLike:
    """
    Advenct points from quasi random uniform distribution.

    Parameters
    ----------
    field : ArrayLike
        Field array with shape T x D x (Z) x Y x X
    img_shape : Tuple[int, ...]
        Must be D-dimensional.
    n_samples : int
        Number of samples.
    invert : bool
        When true flow is multiplied by -1, resulting in reversal of the flow.
    device : Optional[th.device]
        Torch device, by default uses last GPU if available or mps.

    Returns
    -------
    ArrayLike
        Trajectories of sources N x T x D
    """

    if field.shape[1] != len(img_shape):
        raise ValueError(
            f"Field dimension {field.shape[1]} does not match image shape {len(img_shape)}."
        )

    if device is None:
        device = torch_default_device()

    sources = th.quasirandom.SobolEngine(field.shape[1], seed=42).draw(n_samples)
    sources = sources.to(device)
    sources *= th.as_tensor(img_shape, device=device)

    trajectories = advenct_field(field, sources, img_shape, invert=invert)

    trajectories = th.minimum(trajectories, th.as_tensor(img_shape, device=device))
    trajectories = th.maximum(trajectories, th.zeros(len(img_shape), device=device))

    return trajectories


def trajectories_to_tracks(trajectories: th.Tensor) -> np.ndarray:
    """Converts trajectories to napari tracks format.

    Parameters
    ----------
    trajectories : th.Tensor
        Input N x T x D trajectories.

    Returns
    -------
    np.ndarray
        Napari tracks (N x T) x (2 + D) array.
    """
    trajectories = trajectories.cpu().numpy()
    N, T, D = trajectories.shape

    track_ids = np.repeat(np.arange(N), T)[..., None]
    time_pts = np.tile(np.arange(T), N)[..., None]
    coordinates = trajectories.reshape(-1, D)

    return np.concatenate((track_ids, time_pts, coordinates), axis=1)


def _to_tensor(
    image: ArrayLike,
    channel_axis: Optional[int],
    device: th.device,
) -> th.Tensor:
    """Expands image to 4D tensor with channel axis first and converts to tensor

    Parameters
    ----------
    image : ArrayLike
        Image array.
    channel_axis : Optional[int]
        Channel axis, if None image is expanded to 4D tensor with channel axis first.
    device : th.device
        Torch device.

    Returns
    -------
    th.Tensor
        (C, (Z), Y, X)-tensor.
    """
    image = np.asarray(image)

    if channel_axis is None:
        image = image[np.newaxis, ...]

    elif channel_axis > 0:
        image = np.transpose(
            image,
            (channel_axis, *range(channel_axis), *range(channel_axis + 1, image.ndim)),
        )

    return th.as_tensor(image.astype(np.float32), device=device)


def timelapse_flow(
    images: ArrayLike,
    store_or_path: Optional[StoreLike] = None,
    chunks: Optional[Tuple[int]] = None,
    channel_axis: Optional[int] = None,
    im_factor: int = 4,
    grid_factor: int = 4,
    num_iterations: int = 1000,
    lr: float = 1e-2,
    n_scales: int = 3,
    device: Optional[th.device] = None,
) -> zarr.Array:
    """Compute vector field from timelapse.

    Parameters
    ----------
    images : ArrayLike
        Timelapse images shape as (T, ...).
    store_or_path : Optional[StoreLike], optional
        Zarr storage or output path, if not provided a temporary store is used.
    chunks : Optional[Tuple[int]], optional
        Chunk size, if not provided it chunks time with 1 and the spatial dimensions as big as possible.
    channel_axis : Optional[int], optional
        Channel axis EXCLUDING TIME (first axis), e.g (T, C, Y, X) would have `channel_axis=0`.
        If not provided assumes first axis after time.
    im_factor : int, optional
        Image space down scaling factor, by default 4.
    grid_factor : int, optional
        Grid space down scaling factor, by default 4.
        Grid dimensions will be divided by both `im_factor` and `grid_factor`.
    num_iterations : int, optional
        Number of gradient descent iterations, by default 2000.
    lr : float, optional
        Learning rate (gradient descent step), by default 1e-4
    n_scales : int, optional
        Number of scales used for multi-scale optimization, by default 3.
    device : Optional[th.device], optional
        Torch device, by default uses last GPU if available or mps.

    Returns
    -------
    zarr.Array
        Vector field array with shape (T, D, (Z), Y, X).
    """
    if device is None:
        device = torch_default_device()

    imgs_shape = list(images.shape[1:])
    if channel_axis is not None:
        imgs_shape.pop(channel_axis)

    resized_shape = [max(1, int(s // im_factor)) for s in imgs_shape]
    shape = (images.shape[0], len(imgs_shape), *resized_shape)  # (T, D, (Z), Y, X)
    dtype = np.float16

    output = create_zarr(
        shape,
        dtype=dtype,
        store_or_path=store_or_path,
        chunks=chunks,
        default_store_type=None,
    )

    target = _to_tensor(images[0], channel_axis, device)

    for t in tqdm(range(1, len(images)), "Computing flow"):
        source = _to_tensor(images[t], channel_axis, device)

        output[t] = (
            flow_field(
                source,
                target,
                im_factor=im_factor,
                grid_factor=grid_factor,
                lr=lr,
                n_scales=n_scales,
                num_iterations=num_iterations,
            )
            .cpu()
            .numpy()
        )

        target = source

    return output


def add_flow(
    config: MainConfig,
    vector_field: Union[ArrayLike, Sequence[ArrayLike]],
) -> None:
    """
    Adds vector field (coordinate shift) data into nodes.
    If there are fewer vector fields than dimensions, the last dimensions from (z,y,x) have priority.
    For example, if 2 vector fields are provided for a 3D data, only (y, x) are updated.
    Vector field shape, except `t`, can be different from the original image.
    When this happens, the indexing is done mapping the position and rounding.

    Parameters
    ----------
    data_config : DataConfig
        Data configuration parameters.
    vector_field : Sequence[ArrayLike]
        Vector field arrays. Each array per coordinate or a single (T, D, (Z), Y, X)-array.
    """
    LOG.info("Adding shift (vector field) to nodes")

    if isinstance(vector_field, Sequence):
        assert all(v.shape == vector_field[0].shape for v in vector_field)
        ndim = vector_field[0].ndim
        nvecs = len(vector_field)
        vec_shape = np.asarray(vector_field[0].shape)
        is_sequence = True
    else:
        ndim = vector_field.ndim - 1
        nvecs = vector_field.shape[1]
        vec_shape = np.asarray([vector_field.shape[0], *vector_field.shape[2:]])
        is_sequence = False

    shape = np.asarray(config.data_config.metadata["shape"])
    LOG.info(f"Image shape {shape}, ndim={ndim}")
    LOG.info(f"Vec. shape {shape}")
    LOG.info(f"Num. Vec. {nvecs}")

    if len(shape) != ndim:
        raise ValueError(
            "Original data shape and vector field must have same number of dimensions (ignoring channels)."
            f" Found {len(shape)} and {ndim}."
        )

    if np.any(np.abs(vector_field[0]) > 1):
        raise ValueError(
            "Vector field values must be normalized. "
            f"Found {vector_field[0].min()} and {vector_field[0].max()}."
        )

    columns = ["x_shift", "y_shift", "z_shift"]
    coords_scaling = (vec_shape[1:] - 1) / (shape[1:] - 1)
    coordinate_columns = ["z", "y", "x"][-len(coords_scaling) :]
    vec_index_iterator = list(reversed(range(nvecs)))
    # vec_scaling varies depending on the number of vector field and image dimensions
    vec_scaling = np.asarray(shape[1 + len(coords_scaling) - nvecs :])[::-1]

    LOG.info(f"Coordinate_columns {coordinate_columns}")
    LOG.info(f"Vector field scaling of {vec_scaling}")

    engine = sqla.create_engine(config.data_config.database_path)

    for t in tqdm(range(shape[0])):
        with Session(engine) as session:
            query = session.query(NodeDB.id, NodeDB.z, NodeDB.y, NodeDB.x).where(
                NodeDB.t == t
            )
            df = pd.read_sql_query(query.statement, session.bind, index_col="id")

        if len(df) == 0:
            LOG.warning(f"No node found at time point {t}.")
            continue

        coords = df[coordinate_columns].to_numpy()
        coords = np.round(coords * coords_scaling).astype(int)
        coords = np.minimum(
            np.maximum(0, coords), vec_shape[1:] - 1
        )  # truncating boundary
        coords = tuple(coords.T)

        # default value
        df[columns] = 0.0
        # reversed because z could be missing
        for v, colname in zip(vec_index_iterator, columns):
            if is_sequence:
                df[colname] = np.asarray(vector_field[v][t])[
                    coords
                ]  # asarray due lazy loading formats (e.g. dask)
            else:
                df[colname] = np.asarray(vector_field[t, v])[coords]

        # vector field is between -0.5 to 0.5, so it's scaled to the original image size
        # ours is the torch convection divided by 2.0
        # reference
        # https://pytorch.org/docs/stable/generated/torch.nn.functional.grid_sample.html
        df[columns[:nvecs]] *= vec_scaling  # x, y, z

        df["node_id"] = df.index
        with Session(engine) as session:
            statement = (
                sqla.update(NodeDB)
                .where(NodeDB.id == sqla.bindparam("node_id"))
                .values(
                    z_shift=sqla.bindparam("z_shift"),
                    y_shift=sqla.bindparam("y_shift"),
                    x_shift=sqla.bindparam("x_shift"),
                )
            )
            session.connection().execute(
                statement,
                df[["node_id"] + columns].to_dict("records"),
                execution_options={"synchronize_session": False},
            )
            session.commit()



import logging
from typing import Dict, Optional, Sequence

import napari
import numpy as np
import sqlalchemy as sqla
from magicgui.widgets import ComboBox, Container, FloatSlider, Label
from scipy import interpolate

from ultrack.config import MainConfig
from ultrack.core.database import GTLinkDB, NodeDB
from ultrack.utils.ultrack_array import UltrackArray
from ultrack.widgets.ultrackwidget import UltrackWidget

logging.basicConfig()
logging.getLogger("sqlachemy.engine").setLevel(logging.INFO)

LOG = logging.getLogger(__name__)
# LOG.setLevel(logging.INFO)


class HierarchyVizWidget(Container):
    HIER_LAYER_NAME = "Ultrack Hierarchy"

    def __init__(
        self,
        viewer: napari.Viewer,
        config: Optional[MainConfig] = None,
        **kwargs,
    ) -> None:
        """
        Initialize the HierarchyVizWidget.

        Parameters
        ----------
        viewer : napari.Viewer
            The napari viewer instance.
        config : MainConfig of Ultrack
            if not provided, config will be taken from UltrackWidget
        **kwargs : dict
            Additional keyword arguments to pass to the UltrackArray constructor.
        """
        super().__init__(layout="vertical")

        self._viewer = viewer
        title_label = Label(value="<h2>Hierarchy Viz. Widget</h2>")
        self.append(title_label)

        self.config = config
        self._ultrack_array = UltrackArray(self.config, **kwargs)

        # TODO: might need to be updated when the slider is changed
        self._mapping = self._create_mapping()

        # Create area threshold container with fixed width and margins
        self._area_widget_container = Container(layout="horizontal", label="Area")
        self.append(self._area_widget_container)

        # Configure slider label
        self._slider_label = Label(label="0")
        self._slider_label.native.setFixedWidth(25)

        # Configure area threshold slider
        self._area_threshold_w = FloatSlider(
            min=0,
            max=1,
            readout=False,
            value=0.5,
            tooltip="Scroll to change the threshold and transverse the hierarchy",
        )
        self._area_threshold_w.changed.connect(self._slider_update)

        self._area_widget_container.append(self._area_threshold_w)
        self._area_widget_container.append(self._slider_label)

        # Configure node attribute combo box
        self._node_attribute_w = ComboBox(
            label="Node Attribute",
            choices=list(self._node_attributes.keys()),
            value="id",
        )
        self._node_attribute_w.changed.connect(self._on_node_attribute_changed)
        self.append(self._node_attribute_w)

        self._add_ultrack_array()

        self._viewer.layers[self.HIER_LAYER_NAME].refresh()

    def _reset_node_attribute_choices(self) -> None:
        engine = sqla.create_engine(self.config.data_config.database_path)
        inspector = sqla.inspect(engine)

        self._node_attributes: Dict[str, sqla.engine.interfaces.ReflectedColumn] = {
            column["name"]: getattr(NodeDB, column["name"])
            for column in inspector.get_columns(NodeDB.__table__.name)
            if getattr(NodeDB, column["name"]) != NodeDB.pickle
        }

        # Add GTLinkDB attributes if the table exists
        if inspector.has_table(GTLinkDB.__tablename__):
            self._node_attributes |= {
                "gt_id": GTLinkDB.target_id,
                "gt_weight": GTLinkDB.weight,
            }

        if hasattr(self, "_node_attribute_w"):
            self._node_attribute_w.choices = list(self._node_attributes.keys())

    @property
    def config(self) -> MainConfig:
        return self._config

    @config.setter
    def config(self, value: Optional[MainConfig]) -> None:
        if value is None:
            self._config = self._get_config()
        else:
            self._config = value

        # these widgets are created after the config is set for the first time
        self._reset_node_attribute_choices()

        if hasattr(self, "_ultrack_array"):
            self._ultrack_array.config = self.config

    def _add_ultrack_array(self) -> None:
        if self.HIER_LAYER_NAME in self._viewer.layers:
            self._viewer.layers.remove(self.HIER_LAYER_NAME)

        scale = self.config.data_config.metadata.get("scale", [1, 1])

        try:
            layer = self._viewer.add_labels(
                self._ultrack_array,
                name=self.HIER_LAYER_NAME,
                scale=scale,
            )
            layer.refresh()
        except TypeError:
            layer = self._viewer.add_image(
                self._ultrack_array,
                name=self.HIER_LAYER_NAME,
                scale=scale,
                colormap="magma",
            )
            layer.reset_contrast_limits()

    def _on_node_attribute_changed(self, value: str) -> None:
        self._ultrack_array.node_attribute = self._node_attributes.get(value, NodeDB.id)
        self._add_ultrack_array()

    @property
    def _shape(self) -> Sequence[int]:
        return self.config.metadata.get("shape", [])

    def _slider_update(self, value: float) -> None:
        mapped_value = self._mapping(value)
        LOG.info("value %d mapped_value: %d", value, mapped_value)
        self._ultrack_array._filter_value = int(mapped_value)
        self._slider_label.label = str(int(mapped_value))
        self._viewer.layers[self.HIER_LAYER_NAME].refresh()

    def _create_mapping(self) -> interpolate.interp1d:
        """
        Creates a pseudo-linear mapping from U[0,1] to full range of number of pixels
            num_pixels = mapping([0,1])
        """
        LOG.info("Creating mapping")
        current_time = int(
            self._viewer.dims.point[0]
        )  # we assume that time is the first dim
        num_pixels_list = self._ultrack_array.get_tp_num_pixels(
            timeStart=current_time, timeStop=current_time
        )
        num_pixels_list.append(self._ultrack_array.minmax[0])
        num_pixels_list.append(self._ultrack_array.minmax[1])
        num_pixels_list.sort()

        x_vec = np.linspace(0, 1, len(num_pixels_list))
        y_vec = np.array(num_pixels_list)
        mapping = interpolate.interp1d(x_vec, y_vec)

        LOG.info("Mapping created")

        return mapping

    def _get_config(self) -> MainConfig:
        """
        Gets config from the Ultrack widget
        """
        ultrack_widget = UltrackWidget.find_ultrack_widget(self._viewer)
        if ultrack_widget is None:
            raise TypeError(
                "config not provided and was not found within ultrack widget"
            )

        return ultrack_widget._data_forms.get_config()



from pathlib import Path
from typing import Literal

import click
from cloudpickle import load

from ultrack.cli.utils import config_option, persistence_option
from ultrack.config import MainConfig
from ultrack.ml.classification import predict_links_prob, predict_nodes_prob


@click.command("add_probs")
@click.argument("classif_pickle_path", type=click.Path(exists=True, path_type=Path))
@click.option(
    "--var",
    type=click.Choice(["nodes", "links", "divisions", "appearances", "disappearances"]),
    default="nodes",
    help="Variable to assign probabilities.",
)
@config_option()
@persistence_option()
def add_probs_cli(
    classif_pickle_path: Path,
    var: Literal["nodes", "links", "divisions", "appearances", "disappearances"],
    config: MainConfig,
    persistence: bool,
) -> None:
    """Predicts and adds nodes' probabilities to the database."""

    with open(classif_pickle_path, "rb") as f:
        classifier = load(f)
        if isinstance(classifier, dict):
            classifier = classifier[var]

    if var == "nodes":
        predict_nodes_prob(config, classifier, persistence_features=persistence)
    elif var == "links":
        predict_links_prob(config, classifier, persistence_features=persistence)
    else:
        # TODO add edges and other probabilities
        raise NotImplementedError(f"Variable {var} not implemented.")



from typing import List

import numpy as np
import pytest

from ultrack.core.segmentation.node import intersects


@pytest.mark.parametrize(
    "bbox1,bbox2,solution",
    [
        ((0, 0, 5, 5), (1, 1, 2, 2), True),
        ((1, 1, 2, 2), (0, 0, 5, 5), True),
        ((2, 2, 2, 6, 6, 6), (1, 1, 1, 4, 4, 4), True),
        ((5, 5, 7, 7), (3, 2, 6, 4), False),
    ],
)
def test_bbox_intersection(bbox1: List[int], bbox2: List[int], solution: bool) -> None:
    bbox1, bbox2 = np.asarray(bbox1), np.asarray(bbox2)
    assert intersects(bbox1, bbox2) == solution



from typing import Callable

import numpy as np
import pandas as pd
from napari.viewer import ViewerModel

from ultrack.utils.constants import NO_PARENT
from ultrack.validation.link_validation import Annotation, LinkValidation


def test_link_validation(
    make_napari_viewer: Callable[[], ViewerModel],
) -> None:

    tracks_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "parent_id": [NO_PARENT, 0, 1, 2, 1, 4],
            "track_id": [1, 1, 2, 2, 3, 3],
            "parent_track_id": [NO_PARENT, NO_PARENT, 1, 1, 1, 1],
            "t": [0, 1, 2, 3, 2, 3],
            "z": [0, 0, 0, 0, 0, 0],
            "y": [10, 20, 30, 10, 20, 10],
            "x": [10, 20, 30, 10, 20, 10],
        }
    )

    image = np.random.randint(255, size=(5, 64, 64))

    viewer = make_napari_viewer()
    widget = LinkValidation(
        image,
        tracks_df=tracks_df,
        viewer=viewer,
    )
    widget.btn_done.clicked.connect(viewer.close)

    assert not widget.btn_undo.isEnabled()
    widget.btn_correct.click()
    assert widget.btn_undo.isEnabled()

    widget.btn_skip.click()

    widget.btn_undo.click()
    widget.btn_undo.click()

    assert np.all(widget.annotations["annotation"] == Annotation.UNLABELED)
    assert not widget.btn_undo.isEnabled()

    widget.btn_incorrect.click()
    assert widget.btn_undo.isEnabled()

    # reseting
    widget.btn_undo.click()

    # add annotations
    widget.annotations = pd.DataFrame(
        {"id": [1, 2], "annotation": [Annotation.CORRECT, Annotation.INCORRECT]}
    )
    assert widget.current_idx == 2

    widget.btn_correct.click()



from typing import Optional, Tuple

import numpy as np
import pandas as pd
import zarr
from zarr.storage import StoreLike

from ultrack.config.config import MainConfig
from ultrack.core.export.utils import export_segmentation_generic
from ultrack.utils.array import create_zarr, large_chunk_size


def tracks_to_zarr(
    config: MainConfig,
    tracks_df: pd.DataFrame,
    store_or_path: Optional[StoreLike] = None,
    chunks: Optional[Tuple[int]] = None,
    overwrite: bool = False,
) -> zarr.Array:
    """
    Exports segmentations masks to zarr array, `track_df` assign the `track_id` to their respective segments.
    By changing the `store` this function can be used to write zarr arrays into disk.

    Parameters
    ----------
    config : MainConfig
        Configuration parameters.
    tracks_df : pd.DataFrame
        Tracks dataframe, must have `track_id` column and be indexed by node id.
    store_or_path : Optional[StoreLike], optional
        Zarr storage or output path, if not provided a temporary store is used.
    chunks : Optional[Tuple[int]], optional
        Chunk size, if not provided it chunks time with 1 and the spatial dimensions as big as possible.
    overwrite : bool, optional
        If True, overwrites existing zarr array.

    Returns
    -------
    zarr.Array
        Output zarr array.
    """

    shape = config.data_config.metadata["shape"]
    dtype = np.int32

    if (
        isinstance(store_or_path, zarr.storage.MemoryStore)
        and config.data_config.n_workers > 1
    ):
        raise ValueError(
            "zarr.storage.MemoryStore and multiple workers are not allowed. "
            f"Found {config.data_config.n_workers} workers in `data_config`."
        )

    if chunks is None:
        chunks = large_chunk_size(shape, dtype=dtype)

    array = create_zarr(
        shape,
        dtype=dtype,
        store_or_path=store_or_path,
        chunks=chunks,
        default_store_type=None,
        overwrite=overwrite,
    )

    export_segmentation_generic(config.data_config, tracks_df, array.__setitem__)
    return array



import logging
from typing import Optional

import napari
import numpy as np
from magicgui.widgets import Container, PushButton, create_widget
from napari.layers import Tracks

from ultrack.tracks.sorting import sort_trees_by_length
from ultrack.widgets.utils import wait_cursor

LOG = logging.getLogger(__name__)


class TrackInspectionWidget(Container):
    def __init__(self, viewer: napari.Viewer) -> None:
        super().__init__()

        self._viewer = viewer
        self._current_track_layer = None

        self._sorted_tracks = []
        self._tree_index = 0

        self._tracks_layer_w = create_widget(annotation=Tracks)
        self._tracks_layer_w.changed.connect(self._on_layer_change)

        self._next_btn = PushButton(text="Next", enabled=False)
        self._next_btn.changed.connect(self._on_next)

        self._prev_btn = PushButton(text="Prev", enabled=False)
        self._prev_btn.changed.connect(self._on_prev)

        self.append(self._tracks_layer_w)
        self.append(self._next_btn)
        self.append(self._prev_btn)

    @property
    def tree_index(self) -> int:
        return self._tree_index

    @tree_index.setter
    def tree_index(self, value: int) -> int:
        if 0 <= value < len(self._sorted_tracks):
            self._tree_index = value
            self._prev_btn.enabled = value > 0
            self._next_btn.enabled = value < len(self._sorted_tracks) - 1
            if self._current_track_layer is not None:
                subtree = self._sorted_tracks[self._tree_index].to_numpy()
                self._current_track_layer.data = subtree
                self._current_track_layer.name = f"subtree {int(subtree[0, 0])}"

    def _on_next(self) -> None:
        self.tree_index += 1

    def _on_prev(self) -> None:
        self.tree_index -= 1

    def _on_layer_change(self, layer: Optional[Tracks]) -> None:
        LOG.info(f"Track inspection layer update: {layer}")
        if self._current_track_layer is not None:
            self._viewer.layers.remove(self._current_track_layer.name)

        if layer is None:
            return

        with wait_cursor():
            self._sorted_tracks = sort_trees_by_length(layer.data, layer.graph)

        self._current_track_layer = self._viewer.add_tracks(
            np.zeros((2, 4), dtype=float),
            scale=layer.scale[-3:],
            translate=layer.translate[-3:],
            colormap="twilight",
        )
        self.tree_index = 0



from pathlib import Path
from typing import Callable

import numpy as np
import pandas as pd
import pytest
from napari.plugins import _initialize_plugins
from napari.viewer import ViewerModel

from ultrack.reader.napari_reader import napari_get_reader
from ultrack.utils.constants import NO_PARENT


@pytest.fixture
def tracks_df(n_nodes: int = 10) -> pd.DataFrame:
    coordinates = np.random.rand(n_nodes, 3)
    tracks_data = np.zeros((n_nodes, 5))
    tracks_data[:, 2:] = coordinates

    tracks_data[: n_nodes // 2, 0] = 1
    tracks_data[n_nodes // 2 :, 0] = 2
    tracks_data[: n_nodes // 2, 1] = np.arange(n_nodes // 2)
    tracks_data[n_nodes // 2 :, 1] = np.arange(n_nodes - n_nodes // 2)

    return pd.DataFrame(tracks_data, columns=["track_id", "t", "z", "y", "x"])


@pytest.mark.parametrize("file_ext", ["csv", "parquet"])
def test_reader(tracks_df: pd.DataFrame, tmp_path: Path, file_ext: str):
    reader = napari_get_reader(f"tracks.{file_ext}")
    assert reader is None

    path = tmp_path / f"good_tracks.{file_ext}"
    tracks_df["node_id"] = np.arange(len(tracks_df)) + 1
    tracks_df["labels"] = np.random.randint(2, size=len(tracks_df))
    if file_ext == "csv":
        tracks_df.to_csv(path, index=False)
    else:
        tracks_df.to_parquet(path)

    reader = napari_get_reader(path)
    assert callable(reader)

    data, kwargs, type = reader(path)[0]
    assert type == "tracks"

    props = kwargs["features"]

    assert np.allclose(props["node_id"], tracks_df["node_id"])
    assert np.allclose(props["labels"], tracks_df["labels"])
    assert np.allclose(data, tracks_df[["track_id", "t", "z", "y", "x"]])


@pytest.mark.parametrize("file_ext", ["csv", "parquet"])
def test_reader_2d(tracks_df: pd.DataFrame, tmp_path: Path, file_ext: str):
    reader = napari_get_reader(f"tracks.{file_ext}")
    assert reader is None

    path = tmp_path / f"good_tracks.{file_ext}"
    tracks_df = tracks_df.drop(columns=["z"])
    if file_ext == "csv":
        tracks_df.to_csv(path, index=False)
    else:
        tracks_df.to_parquet(path)

    reader = napari_get_reader(path)
    assert callable(reader)

    data, _, type = reader(path)[0]
    assert type == "tracks"

    assert np.allclose(data, tracks_df[["track_id", "t", "y", "x"]])


@pytest.mark.parametrize("file_ext", ["csv", "parquet"])
def test_reader_with_lineage(tmp_path: Path, file_ext: str):
    tracks_df = pd.DataFrame(
        {
            "track_id": [1, 1, 2, 3],
            "parent_track_id": [NO_PARENT, NO_PARENT, 1, 1],
            "t": [0, 1, 2, 2],
            "z": [0, 0, 0, 0],
            "y": [10, 20, 30, 10],
            "x": [1, 2, 3, 1],
        }
    )

    path = tmp_path / f"tracks.{file_ext}"
    if file_ext == "csv":
        tracks_df.to_csv(path, index=False)
    else:
        tracks_df.to_parquet(path)

    reader = napari_get_reader(path)
    assert callable(reader)

    data, kwargs, type = reader(path)[0]
    assert type == "tracks"
    assert "graph" in kwargs
    assert kwargs["graph"] == {2: 1, 3: 1}

    assert np.allclose(data, tracks_df[["track_id", "t", "z", "y", "x"]])


def test_non_existing_track():
    reader = napari_get_reader("tracks.csv")
    assert reader is None


@pytest.mark.parametrize("file_ext", ["csv", "parquet"])
def test_wrong_columns_track(tracks_df: pd.DataFrame, tmp_path: Path, file_ext: str):
    reader = napari_get_reader(f"tracks.{file_ext}")
    assert reader is None

    path = tmp_path / f"bad_tracks.{file_ext}"
    tracks_df = tracks_df.rename(columns={"track_id": "id"})
    if file_ext == "csv":
        tracks_df.to_csv(path, index=False)
    else:
        tracks_df.to_parquet(path)

    reader = napari_get_reader(path)
    assert reader is None


@pytest.mark.parametrize("file_ext", ["csv", "parquet"])
def test_napari_viewer_open_tracks(
    make_napari_viewer: Callable[[], ViewerModel],
    tracks_df: pd.DataFrame,
    tmp_path: Path,
    file_ext: str,
) -> None:

    _initialize_plugins()

    path = tmp_path / f"tracks.{file_ext}"
    if file_ext == "csv":
        tracks_df.to_csv(path, index=False)
    else:
        tracks_df.to_parquet(path)

    viewer = make_napari_viewer()
    viewer.open(path, plugin="ultrack")



import logging
import uuid
from pathlib import Path
from typing import Literal, Optional

import mip
import numpy as np
import pandas as pd
from numpy.typing import ArrayLike
from skimage.util._map_array import ArrayMap

from ultrack.config.config import TrackingConfig
from ultrack.core.solve.solver.base_solver import BaseSolver
from ultrack.utils.array import assert_same_length
from ultrack.utils.constants import NO_PARENT

LOG = logging.getLogger(__name__)

_KEY_TO_SOLVER_NAME = {
    "CBC": "Coin-OR Branch and Cut",
    "GRB": "Gurobi",
    "GUROBI": "Gurobi",
}


class MIPSolver(BaseSolver):
    def __init__(
        self,
        config: TrackingConfig,
    ) -> None:
        """Generic mixed-integer programming (MIP) solver for cell-tracking ILP.

        Parameters
        ----------
        config : TrackingConfig
            Tracking configuration parameters.
        """
        self._config = config
        self.reset()

    def reset(self) -> None:
        """Sets model to an empty state."""
        try:
            self._model = mip.Model(
                sense=mip.MAXIMIZE, solver_name=self._config.solver_name
            )
        except mip.exceptions.InterfacingError as e:
            LOG.warning(e)
            self._model = mip.Model(sense=mip.MAXIMIZE, solver_name=mip.CBC)

        print(f"Using {_KEY_TO_SOLVER_NAME[self._model.solver_name]} solver")

        if self._model.solver_name == mip.CBC:
            LOG.warning(
                "Using CBC solver. Consider installing Gurobi for significantly better performance."
            )
            LOG.warning(
                "To install Gurobi, follow the instructions at "
                "https://support.gurobi.com/hc/en-us/articles/360044290292-How-do-I-install-Gurobi-for-Python-"
            )
            LOG.warning(
                "It is free for academic use. "
                "See https://www.gurobi.com/academia/academic-program-and-licenses/"
            )

        self._forward_map = None
        self._backward_map = None
        self._nodes = None
        self._appearances = None
        self._disappearances = None
        self._divisions = None
        self._edges = None
        self._weights = None
        self._setup_model_parameters()

    def _setup_model_parameters(self) -> None:
        """Sets model parameters from configuration file."""
        self._model.max_seconds = self._config.time_limit
        self._model.threads = self._config.n_threads
        self._model.lp_method = self._config.method
        self._model.max_mip_gap = self._config.solution_gap

    def add_nodes(
        self,
        indices: ArrayLike,
        is_first_t: ArrayLike,
        is_last_t: ArrayLike,
        is_border: ArrayLike = False,
        nodes_prob: Optional[ArrayLike] = None,
    ) -> None:
        """Add nodes slack variables to gurobi model.

        Parameters
        ----------
        indices : ArrayLike
            Nodes indices.
        is_first_t : ArrayLike
            Boolean array indicating if it belongs to first time point and it won't receive appearance penalization.
        is_last_t : ArrayLike
            Boolean array indicating if it belongs to last time point and it won't receive disappearance penalization.
        is_border : ArrayLike
            Boolean array indicating if it belongs to the border and it won't receive (dis)apperance penalization.
            Default: False
        nodes_prob: Optional[ArrayLike]
            If provided assigns a node probability score to the objective function.
        """
        if self._nodes is not None:
            raise ValueError("Nodes have already been added.")

        assert_same_length(
            indices=indices,
            is_first_t=is_first_t,
            is_last_t=is_last_t,
            nodes_prob=nodes_prob,
        )

        LOG.info("# %s nodes at starting `t`.", np.sum(is_first_t))
        LOG.info("# %s nodes at last `t`.", np.sum(is_last_t))

        appear_weight = (
            np.logical_not(is_first_t | is_border) * self._config.appear_weight
        )
        disappear_weight = (
            np.logical_not(is_last_t | is_border) * self._config.disappear_weight
        )

        indices = np.asarray(indices, dtype=int)
        self._backward_map = np.array(indices, copy=True)
        self._forward_map = ArrayMap(indices, np.arange(len(indices)))
        size = (len(indices),)

        self._nodes = self._model.add_var_tensor(
            size, name="nodes", var_type=mip.BINARY
        )
        self._appearances = self._model.add_var_tensor(
            size, name="appear", var_type=mip.BINARY
        )
        self._disappearances = self._model.add_var_tensor(
            size, name="disappear", var_type=mip.BINARY
        )
        self._divisions = self._model.add_var_tensor(
            size, name="division", var_type=mip.BINARY
        )

        if nodes_prob is None:
            node_weights = 0
        else:
            nodes_prob = self._config.apply_link_function(np.asarray(nodes_prob))
            node_weights = mip.xsum(nodes_prob * self._nodes)

        self._model.objective = (
            mip.xsum(self._divisions * self._config.division_weight)
            + mip.xsum(self._appearances * appear_weight)
            + mip.xsum(self._disappearances * disappear_weight)
            + node_weights
        )

    def add_edges(
        self, sources: ArrayLike, targets: ArrayLike, weights: ArrayLike
    ) -> None:
        """Add edges to model and applies weights link function from config.

        Parameters
        ----------
        source : ArrayLike
            Array of integers indicating source indices.
        targets : ArrayLike
            Array of integers indicating target indices.
        weights : ArrayLike
            Array of weights, input to the link function.
        """
        if self._edges is not None:
            raise ValueError("Edges have already been added.")

        assert_same_length(sources=sources, targets=targets, weights=weights)

        weights = self._config.apply_link_function(weights.astype(float))

        LOG.info("transformed edge weights %s", weights)

        sources = self._forward_map[np.asarray(sources, dtype=int)]
        targets = self._forward_map[np.asarray(targets, dtype=int)]

        self._edges = self._model.add_var_tensor(
            (len(weights),), name="edges", var_type=mip.BINARY
        )
        self._edges_df = pd.DataFrame(
            np.asarray([sources, targets]).T, columns=["sources", "targets"]
        )

        self._model.objective += mip.xsum(weights * self._edges)

    def set_standard_constraints(self) -> None:
        """Sets standard biological/flow constraints:
        - single incoming node (no fusion);
        - flow conservation (begin and end requires slack variables);
        - divisions only from existing nodes;
        """
        edges_targets = self._edges_df.groupby("targets")
        edges_sources = self._edges_df.groupby("sources")

        for i in range(self._nodes.shape[0]):
            # yes, it's flipped, when sources are fixed we access targets
            try:
                i_sources = edges_targets.get_group(i).index
            except KeyError:
                i_sources = []

            try:
                i_targets = edges_sources.get_group(i).index
            except KeyError:
                i_targets = []

            # single incoming node
            self._model.add_constr(
                mip.xsum(self._edges[i_sources]) + self._appearances[i]
                == self._nodes[i]
            )

            # flow conservation
            self._model.add_constr(
                self._nodes[i] + self._divisions[i]
                == mip.xsum(self._edges[i_targets]) + self._disappearances[i]
            )

            # divisions
            self._model.add_constr(self._nodes[i] >= self._divisions[i])

    def add_overlap_constraints(self, sources: ArrayLike, targets: ArrayLike) -> None:
        """Add constraints such that `source` and `target` can't be present in the same solution.

        Parameters
        ----------
        source : ArrayLike
            Source nodes indices.
        target : ArrayLike
            Target nodes indices.
        """
        sources = self._forward_map[np.asarray(sources, dtype=int)]
        targets = self._forward_map[np.asarray(targets, dtype=int)]

        for i in range(len(sources)):
            self._model.add_constr(
                self._nodes[sources[i]] + self._nodes[targets[i]] <= 1
            )

    def enforce_nodes_solution_value(
        self,
        indices: ArrayLike,
        variable: Literal["appear", "disappear", "division", "node"],
        value: bool,
    ) -> None:
        """Constraints given nodes' variables to 1.

        Parameters
        ----------
        indices : ArrayLike
            Nodes indices.
        variable : str
            Slack variable to constraint.
        value : bool
            Value to constraint to.
        """
        indices = self._forward_map[np.asarray(indices, dtype=int)]

        variable_arr = {
            "appear": self._appearances,
            "disappear": self._disappearances,
            "division": self._divisions,
            "node": self._nodes,
        }[variable]

        if value:
            for i in indices:
                self._model.add_constr(variable_arr[i] >= 1)
        else:
            for i in indices:
                self._model.add_constr(variable_arr[i] <= 0)

    def enforce_edges_solution_value(
        self,
        sources: ArrayLike,
        targets: ArrayLike,
        value: bool,
    ) -> None:
        """Constraints given nodes' variables to 1.

        Parameters
        ----------
        source : ArrayLike
            Array of integers indicating source indices (at time T - 1).
        targets : ArrayLike
            Array of integers indicating target indices (at time T).
        value : bool
            Value to constraint to.
        """
        if self._edges_df is None:
            raise ValueError("Edges must be added before enforcing their value.")

        sources = self._forward_map[np.asarray(sources, dtype=int)]
        targets = self._forward_map[np.asarray(targets, dtype=int)]

        # saving indices
        df = self._edges_df.reset_index()

        match = pd.merge(
            df,
            pd.DataFrame({"sources": sources, "targets": targets}),
            on=["sources", "targets"],
            validate="1:1",
        )
        if len(match) != len(sources):
            raise ValueError(
                f"{len(sources) - len(match)} edges were not found at `enforce_edges_solution_value`."
            )

        if value:
            for i in match["index"]:
                self._model.add_constr(self._edges[i] >= 1)
        else:
            for i in match["index"]:
                self._model.add_constr(self._edges[i] <= 0)

    def set_nodes_sum(self, indices: ArrayLike, total_sum: int) -> None:
        """Set indices sum to total_sum as constraint.

        sum_i nodes[i] = total_sum

        Parameters
        ----------
        indices : ArrayLike
            Nodes indices.
        total_sum : int
            Total sum of nodes' variables.
        """
        indices = self._forward_map[np.asarray(indices, dtype=int)]
        self._model.add_constr(mip.xsum([self._nodes[i] for i in indices]) == total_sum)

    def _set_solution_guess(self) -> None:
        # TODO
        pass

    def optimize(self) -> float:
        """Optimizes MIP model."""
        self._model.optimize()
        return self._model.objective_value

    def solution(self) -> pd.DataFrame:
        """Returns the nodes present on the solution.

        Returns
        -------
        pd.DataFrame
            Dataframe indexed by nodes as indices and their parent (NA if orphan).
        """
        if self._model.status == mip.OptimizationStatus.FEASIBLE:
            LOG.warning(
                f"Solver status {self._model.status}. Search interrupted before conclusion."
            )

        elif self._model.status == mip.OptimizationStatus.INFEASIBLE:
            model_path = Path(".") / f"{uuid.uuid4()}_tracking.lp"
            self._model.write(str(model_path.absolute()))
            raise ValueError(
                f"Infeasible solution found. Exported tracking LP to {model_path.absolute()} for debugging."
            )

        elif self._model.status != mip.OptimizationStatus.OPTIMAL:
            raise ValueError(
                f"Solver must be optimized before returning solution. It had status {self._model.status}"
            )

        nodes = np.asarray(
            [i for i, node in enumerate(self._nodes) if node.x > 0.5], dtype=int
        )
        nodes = self._backward_map[nodes]
        LOG.info("Solution nodes\n%s", nodes)

        if len(nodes) == 0:
            raise ValueError("Something went wrong, nodes solution is empty.")

        nodes = pd.DataFrame(
            data=NO_PARENT,
            index=nodes,
            columns=["parent_id"],
        )

        edges_solution = np.asarray(
            [i for i, edge in enumerate(self._edges) if edge.x > 0.5], dtype=int
        )
        edges = self._backward_map[self._edges_df.loc[edges_solution].values]

        LOG.info("Solution edges\n%s", edges)

        if len(edges) == 0:
            raise ValueError("Something went wrong, edges solution is empty")

        edges = pd.DataFrame(
            data=edges[:, 0],
            index=edges[:, 1],
            columns=["parent_id"],
        )

        nodes.update(edges)

        return nodes



from abc import abstractmethod
from pathlib import Path

import napari
from magicgui.widgets import FileEdit

from ultrack.config import DataConfig, load_config
from ultrack.widgets.ultrackwidget._legacy.baseconfigwidget import BaseConfigWidget


class GenericDataWidget(BaseConfigWidget):
    def __init__(self, viewer: napari.Viewer, label: str) -> None:
        self._viewer = viewer
        super().__init__(DataConfig(), label)

        self._config_loader_w = FileEdit(
            filter="*toml",
            label="Config. Path",
            value=None,
        )
        self.append(self._config_loader_w)

    def _setup_widgets(self) -> None:
        pass

    def _on_config_loaded(self, value: Path) -> None:
        if value.exists() and value.is_file():
            self.config = load_config(value).data_config

    @BaseConfigWidget.config.setter
    def config(self, value: DataConfig) -> None:
        BaseConfigWidget.config.fset(self, value)
        self._on_config_changed()

    @abstractmethod
    def _on_config_changed(self) -> None:
        pass



from qtpy.QtWidgets import QLineEdit

from ultrack.widgets.ultrackwidget.components.number_validators import (
    DoubleOrEmptyValidator,
    IntOrEmptyValidator,
)


class BlankableNumberEdit(QLineEdit):
    def __init__(
        self, parent=None, default=0, minimum=-999999, maximum=999999, dtype=int
    ):
        super().__init__(parent)

        self.dtype = dtype
        if dtype == int:
            self.setValidator(IntOrEmptyValidator(minimum, maximum, self))
        elif dtype == float:
            self.setValidator(DoubleOrEmptyValidator(minimum, maximum, 5, self))
        else:
            raise ValueError("dtype must be int or float")

        self.setText(str(default))

    def getValue(self):
        if self.text().strip() == "":
            return None
        return self.dtype(self.text())



import numpy as np
import pandas as pd
import pytest

from ultrack.config.config import MainConfig
from ultrack.core.solve.solver.mip_solver import MIPSolver
from ultrack.core.solve.sqltracking import _check_inside_border


@pytest.mark.parametrize(
    "config_content",
    [
        {
            "tracking.appear_weight": -0.25,
            "tracking.disappear_weight": -1.0,
            "tracking.division_weight": -0.5,
            "tracking.link_function": "identity",
            "tracking.bias": 0,
        }
    ],
    indirect=True,
)
def test_solvers_optimize(config_instance: MainConfig) -> None:
    """
    This demo builds a very simple graph with 7 nodes and a single overlap constraint (2,5)
    and a two possible divisions on 2 and 6.

    Edge -C- denotes contraint.

    Graph:

    1 - 0.5 - 2 - 0.5 - 3 - 0.5 - 4
              |  \\             /      \\ due linting software
              C   1.0      0.9
              |       \\  /
              5 - 0.5 - 6 - 0.7 - 7

    Solution:

    1 - 0.5 - 2 - 0.5 - 3 - 0.5 - 4
                \
                  1.0
                     \
                       6 - 0.7 - 7

    Result: 0.5 + 0.5 + 1.0 + 0.7 - division_weight
    """
    solver = MIPSolver(config_instance.tracking_config)

    nodes = np.array([1, 2, 3, 4, 5, 6, 7])
    is_first = np.array([1, 0, 0, 0, 0, 0, 0], dtype=bool)
    is_last = np.array([0, 0, 0, 1, 0, 0, 1], dtype=bool)

    solver.add_nodes(nodes, is_first, is_last)

    edges = np.array([[1, 2], [2, 3], [2, 6], [3, 4], [5, 6], [6, 4], [6, 7]])

    weights = np.array([0.5, 0.5, 1.0, 0.5, 0.5, 0.9, 0.7])
    solver.add_edges(edges[:, 0], edges[:, 1], weights)

    solver.set_standard_constraints()

    solver.add_overlap_constraints([2], [5])

    objective = solver.optimize()
    solution = solver.solution()

    expected_solution = pd.DataFrame(
        data=[pd.NA, 1, 2, 3, 2, 6],
        index=[1, 2, 3, 4, 6, 7],
        columns=["parent_id"],
        dtype=pd.Int64Dtype(),
    )
    expected_edges = np.array([1, 1, 1, 1, 0, 0, 1], dtype=bool)

    assert solution.shape == expected_solution.shape
    assert np.all(solution.index.isin(expected_solution.index))
    assert np.all(
        expected_solution.loc[solution.index, "parent_id"] == solution["parent_id"]
    )
    assert np.allclose(
        objective,
        weights[expected_edges].sum() + config_instance.tracking_config.division_weight,
    )


def test_fixed_nodes_constraint_solver(config_instance: MainConfig) -> None:
    """
    Same graph as before but with a fixed division on node 6.

    Edge -C- denotes contraint.

    Graph:

    1 - 0.5 - 2 - 0.5 - 3 - 0.5 - 4
              |  \\             /      \\ due linting software
              C   1.0      0.9
              |       \\  /
              5 - 0.5 - 6 - 0.7 - 7
                        ^
                        |
            THIS WILL IS FIXED AS DIVISION

    Solution:

    1 - 0.5 - 2                   4
                \\              /
                  1.0      0.9
                     \\  /
                       6 - 0.7 - 7

    Result: 0.5 + 1.0 + 0.7 + 0.9 - division_weight
    """
    solver = MIPSolver(config_instance.tracking_config)

    nodes = np.array([1, 2, 3, 4, 5, 6, 7])
    is_first = np.array([1, 0, 0, 0, 0, 0, 0], dtype=bool)
    is_last = np.array([0, 0, 0, 1, 0, 0, 1], dtype=bool)

    solver.add_nodes(nodes, is_first, is_last)

    edges = np.array([[1, 2], [2, 3], [2, 6], [3, 4], [5, 6], [6, 4], [6, 7]])

    weights = np.array([0.5, 0.5, 1.0, 0.5, 0.5, 0.9, 0.7])
    solver.add_edges(edges[:, 0], edges[:, 1], weights)

    solver.set_standard_constraints()

    solver.add_overlap_constraints([2], [5])

    solver.enforce_nodes_solution_value([6], "division", True)

    objective = solver.optimize()
    solution = solver.solution()

    expected_solution = pd.DataFrame(
        data=[pd.NA, 1, 6, 2, 6],
        index=[1, 2, 4, 6, 7],
        columns=["parent_id"],
        dtype=pd.Int64Dtype(),
    )
    expected_edges = np.array([1, 0, 1, 0, 0, 1, 1], dtype=bool)

    assert solution.shape == expected_solution.shape
    assert np.all(solution.index.isin(expected_solution.index))
    assert np.all(
        expected_solution.loc[solution.index, "parent_id"] == solution["parent_id"]
    )
    assert np.allclose(
        objective,
        weights[expected_edges].sum() + config_instance.tracking_config.division_weight,
    )


def test_solver_with_node_probabilities(config_instance: MainConfig) -> None:
    """
    Edge -C- denotes contraint.

    Graph:

   0.3       0.7       1.0       1.0
    1 - 0.5 - 2 - 0.5 - 3 - 0.5 - 4
              |  \\             /      \\ due linting software
              C   1.0      0.9
              |       \\  /
              5 - 0.5 - 6 - 0.7 - 7
    node w.  0.5       1.0       1.0

    Solution:

   0.3       0.7       1.0       1.0
    1 - 0.5 - 2 - 0.5 - 3 - 0.5 - 4
                 \\
                  1.0
                      \\
                        6 - 0.7 - 7
    node w.  0.5       1.0       1.0

    Result: 0.3 + 0.7 + 1.0 + 1.0 + 1.0 + 1.0 +
            0.5 + 0.5 + 0.5 + 1.0 + 0.7 - division_weight
    """
    solver = MIPSolver(config_instance.tracking_config)

    nodes = np.array([1, 2, 3, 4, 5, 6, 7])
    nodes_probs = np.array([0.3, 0.7, 1.0, 1.0, 0.5, 1.0, 1.0])
    is_first = np.array([1, 0, 0, 0, 0, 0, 0], dtype=bool)
    is_last = np.array([0, 0, 0, 1, 0, 0, 1], dtype=bool)

    solver.add_nodes(nodes, is_first, is_last, nodes_prob=nodes_probs)

    edges = np.array([[1, 2], [2, 3], [2, 6], [3, 4], [5, 6], [6, 4], [6, 7]])

    weights = np.array([0.5, 0.5, 1.0, 0.5, 0.5, 0.9, 0.7])
    solver.add_edges(edges[:, 0], edges[:, 1], weights)

    solver.set_standard_constraints()

    solver.add_overlap_constraints([2], [5])

    objective = solver.optimize()
    solution = solver.solution()

    expected_solution = pd.DataFrame(
        data=[pd.NA, 1, 2, 3, 2, 6],
        index=[1, 2, 3, 4, 6, 7],
        columns=["parent_id"],
        dtype=pd.Int64Dtype(),
    )
    expected_edges = np.array([1, 1, 1, 1, 0, 0, 1], dtype=bool)

    assert solution.shape == expected_solution.shape
    assert np.all(solution.index.isin(expected_solution.index))
    assert np.all(
        expected_solution.loc[solution.index, "parent_id"] == solution["parent_id"]
    )
    assert np.allclose(
        objective,
        nodes_probs[expected_solution.index.to_numpy() - 1].sum()
        + weights[expected_edges].sum()
        + config_instance.tracking_config.division_weight,
    )


def test_fixed_edges_constraint_solver(config_instance: MainConfig) -> None:
    """
    Same graph as before but with a fixed division on node 6.

    Edge -C- denotes contraint.

    Graph:

    1 - 0.5 - 2 - 0.5 - 3 - 0.5 - 4
              |  \\             /      \\ due linting software
              C   1.0      0.9
              |       \\  /
              5 - 0.5 - 6 - 0.7 - 7
                             ^
                             |
                        FORBIDDEN EDGE

    Solution:

    1 - 0.5 - 2                   4
                \\              /
                  1.0      0.9
                     \\  /
                       6

    Result: 0.5 + 1.0 + 0.9
    """
    solver = MIPSolver(config_instance.tracking_config)

    nodes = np.array([1, 2, 3, 4, 5, 6, 7])
    is_first = np.array([1, 0, 0, 0, 0, 0, 0], dtype=bool)
    is_last = np.array([0, 0, 0, 1, 0, 0, 1], dtype=bool)

    solver.add_nodes(nodes, is_first, is_last)

    edges = np.array([[1, 2], [2, 3], [2, 6], [3, 4], [5, 6], [6, 4], [6, 7]])

    weights = np.array([0.5, 0.5, 1.0, 0.5, 0.5, 0.9, 0.7])
    solver.add_edges(edges[:, 0], edges[:, 1], weights)

    solver.set_standard_constraints()

    solver.add_overlap_constraints([2], [5])

    solver.enforce_edges_solution_value([6], [7], False)

    objective = solver.optimize()
    solution = solver.solution()

    expected_solution = pd.DataFrame(
        data=[pd.NA, 1, 6, 2],
        index=[1, 2, 4, 6],
        columns=["parent_id"],
        dtype=pd.Int64Dtype(),
    )
    expected_edges = np.array([1, 0, 1, 0, 0, 1, 0], dtype=bool)

    assert solution.shape == expected_solution.shape
    assert np.all(solution.index.isin(expected_solution.index))
    assert np.all(
        expected_solution.loc[solution.index, "parent_id"] == solution["parent_id"]
    )
    assert np.allclose(objective, weights[expected_edges].sum())


@pytest.mark.parametrize(
    "config_content",
    [
        {
            "tracking.image_border_size": (1, 3, 2),
        },
    ],
    indirect=True,
)
def test_add_nodes_with_is_border(config_instance: MainConfig):
    """
    Test the add_nodes method with the is_border parameter.

    Assuming TZYX shape = (10,10,10,10) and varying image_border_size
    Testing if node is inside the border in all combinations of z, y, x
    """
    solver = MIPSolver(config_instance.tracking_config)

    shape = (10, 10, 10, 10)  # TZYX

    data = {
        "id": np.arange(10),
        "t": np.concatenate([np.zeros(5), np.ones(5)]),
        "z": np.array([0, 0, 0, 0, 5, 0, 5, 9, 5, 9]),
        "y": np.array([5, 2, 2, 0, 5, 5, 2, 9, 5, 5]),
        "x": np.array([5, 5, 5, 0, 1, 1, 1, 9, 5, 9]),
    }
    df = pd.DataFrame(data)
    is_border = _check_inside_border(
        df, config_instance.tracking_config.image_border_size, shape=shape
    )

    # Assert that the is_border Series is as expected
    expected_is_border = np.asarray(
        [True, True, True, True, True, True, True, True, False, True]
    )
    np.testing.assert_array_equal(is_border, expected_is_border)

    # Add nodes to the solver
    solver.add_nodes(
        df["id"],
        # avoiding first and last nodes zero (dist)appearance cost
        np.zeros(10, dtype=bool),
        np.zeros(10, dtype=bool),
        is_border=is_border,
    )

    # Assertions to verify the nodes were added correctly
    assert solver._nodes is not None
    assert solver._appearances is not None
    assert solver._disappearances is not None
    assert solver._divisions is not None

    # Verify that the is_border values are correctly processed
    for idx, border in enumerate(is_border):
        if border:
            # Verify that appearance and disappearance weights are set to 0 for border nodes
            assert solver._appearances[idx].obj == 0
            assert solver._disappearances[idx].obj == 0
        else:
            assert (
                solver._appearances[idx].obj
                == config_instance.tracking_config.appear_weight
            )
            assert (
                solver._disappearances[idx].obj
                == config_instance.tracking_config.disappear_weight
            )



import shutil
from pathlib import Path
from typing import Union

import geff
import numpy as np
import pandas as pd
import sqlalchemy as sqla
import zarr
from geff.core_io import construct_var_len_props, write_arrays
from geff_spec import Axis, PropMetadata
from sqlalchemy.orm import Session

from ultrack.config import MainConfig
from ultrack.core.database import NO_PARENT, LinkDB, NodeDB, OverlapDB


def to_geff(
    config: MainConfig,
    filename: Union[str, Path],
    overwrite: bool = False,
) -> None:
    """
    Export tracks to a geff (Graph Exchange File Format) file.

    Parameters
    ----------
    config : MainConfig
        The configuration object.
    filename : str or Path
        The name of the file to save the tracks to.
    overwrite : bool, optional
        Whether to overwrite the file if it already exists, by default False.

    Raises
    ------
    FileExistsError
        If the file already exists and overwrite is False.
    """
    filename = Path(filename)
    if filename.exists():
        if not overwrite:
            raise FileExistsError(
                f"File {filename} already exists. Set `overwrite=True` to overwrite the file"
            )
        else:
            shutil.rmtree(filename)

    engine = sqla.create_engine(config.data_config.database_path)
    with Session(engine) as session:
        node_stmt = session.query(
            NodeDB.id,
            NodeDB.t,
            NodeDB.parent_id,
            NodeDB.z,
            NodeDB.y,
            NodeDB.x,
            NodeDB.z_shift,
            NodeDB.y_shift,
            NodeDB.x_shift,
            NodeDB.area,
            NodeDB.frontier,
            NodeDB.height,
            NodeDB.selected,
            NodeDB.pickle,
        ).statement
        node_df = pd.read_sql(node_stmt, session.bind, index_col="id")
        node_df["id"] = node_df.index

        edge_stmt = session.query(
            LinkDB.source_id, LinkDB.target_id, LinkDB.weight
        ).statement
        edge_df = pd.read_sql(edge_stmt, session.bind)

        sol_links_df = node_df.loc[
            node_df["selected"] & node_df["parent_id"] != NO_PARENT,
            ["id", "parent_id"],
        ]
        sol_links_df = sol_links_df.rename(
            columns={"parent_id": "source_id", "id": "target_id"},
        )
        sol_links_df["solution"] = True
        edge_df = edge_df.merge(sol_links_df, on=["source_id", "target_id"])
        edge_df["solution"] = edge_df["solution"].fillna(False)

        node_df.rename(columns={"selected": "solution"}, inplace=True)
        node_df.drop(["id", "parent_id"], axis=1, inplace=True)

        overlap_stmt = session.query(
            OverlapDB.node_id,
            OverlapDB.ancestor_id,
        ).statement
        overlap_df = pd.read_sql(overlap_stmt, session.bind)

    node_props_metadata = {
        c: PropMetadata(
            identifier=c,
            dtype=node_df[c].dtype,
        )
        for c in node_df.columns
        if c != "pickle"
    }
    node_props_metadata["mask"] = PropMetadata(
        identifier="mask",
        dtype=np.uint64,
        varlength=True,
    )
    node_props_metadata["bbox"] = PropMetadata(
        identifier="bbox",
        dtype=np.int64,
    )

    edge_ids = edge_df[["source_id", "target_id"]].to_numpy(dtype=np.uint64)
    edge_df = edge_df.drop(columns=["source_id", "target_id"])

    edge_props_metadata = {
        c: PropMetadata(identifier=c, dtype=edge_df[c].dtype) for c in edge_df.columns
    }

    geff_metadata = geff.GeffMetadata(
        directed=True,
        axes=[
            Axis(name="t", type="time"),
            Axis(name="z", type="space"),
            Axis(name="y", type="space"),
            Axis(name="x", type="space"),
        ],
        node_props_metadata=node_props_metadata,
        edge_props_metadata=edge_props_metadata,
    )

    node_props = {
        c: {"values": node_df[c].to_numpy(), "missing": None}
        for c in node_df.columns
        if c != "pickle"
    }
    node_props["mask"] = construct_var_len_props(
        [v.mask.astype(np.uint64) for v in node_df["pickle"]]
    )
    node_props["bbox"] = {
        "values": np.stack([v.bbox for v in node_df["pickle"]]),
        "missing": None,
    }

    write_arrays(
        filename,
        node_ids=node_df.index.to_numpy(dtype=np.uint64),
        node_props=node_props,
        edge_ids=edge_ids,
        edge_props={
            c: {"values": edge_df[c].to_numpy(), "missing": None} for c in edge_df
        },
        metadata=geff_metadata,
    )

    # custom element to geff
    store = zarr.open(filename, mode="a")
    store["overlaps/ids"] = overlap_df[["ancestor_id", "node_id"]].to_numpy(
        dtype=np.uint64
    )
    store.create_group("overlaps/props")



import logging
from pathlib import Path
from typing import Callable, Dict, List, Optional, Set, Tuple

import numpy as np
import pandas as pd
import sqlalchemy as sqla
from numpy.typing import ArrayLike
from scipy.ndimage import generate_binary_structure, grey_dilation, zoom
from scipy.optimize import linear_sum_assignment
from scipy.spatial import KDTree
from scipy.spatial.distance import cdist
from skimage.measure import regionprops
from skimage.segmentation import relabel_sequential
from sqlalchemy.orm import Session
from tifffile import imwrite
from toolz import curry
from tqdm import tqdm

from ultrack.config.config import MainConfig
from ultrack.config.dataconfig import DataConfig
from ultrack.core.database import NodeDB
from ultrack.core.export.utils import (
    export_segmentation_generic,
    filter_nodes_generic,
    solution_dataframe_from_sql,
)
from ultrack.core.segmentation.node import Node, intersects
from ultrack.tracks.graph import (
    add_track_ids_to_tracks_df,
    get_subtree,
    tracks_df_forest,
)
from ultrack.tracks.stats import estimate_drift
from ultrack.utils.constants import NO_PARENT
from ultrack.utils.data import validate_and_overwrite_path

logging.basicConfig()
logging.getLogger("sqlachemy.engine").setLevel(logging.INFO)

LOG = logging.getLogger(__name__)


def ctc_compress_forest(df: pd.DataFrame) -> pd.DataFrame:
    """
    Compress each maximal path into the cell-tracking challenge format.
    Reference: https://public.celltrackingchallenge.net/documents/Naming%20and%20file%20content%20conventions.pdf

    It contains 4 columns L, B, E, P, where:
        - L is unique track label;
        - B is a zero-based track time begin;
        - E is a zero-based track time end;
        - P is parent track index, 0 when parent-less.

    Parameters
    ----------
    df : pd.DataFrame
        Forest dataframe with `track_id` and `parent_track_id` columns.

    Returns
    -------
    pd.DataFrame
        Compressed path forest.
    """
    ctc_df = []
    for track_id, group in df.groupby("track_id"):
        ctc_df.append(
            {
                "L": track_id,
                "B": int(group["t"].min()),
                "E": int(group["t"].max()),
                "P": group["parent_track_id"].iloc[0],
            }
        )

    ctc_df = pd.DataFrame(ctc_df)
    ctc_df.loc[ctc_df["P"] == NO_PARENT, "P"] = 0

    return ctc_df


def stitch_tracks_df(
    graph: Dict[int, List[int]],
    df: pd.DataFrame,
    selected_track_ids: Set[int],
) -> pd.DataFrame:
    """Filters selected tracks and stitches (connects) incomplete tracks to nearby tracks on subsequent time point.

    Parameters
    ----------
    graph : Dict[int, List[int]]
        Forest graph indexed by subtree root.
    df : pd.DataFrame
        Tracks dataframe.
    selected_track_ids : Set[int]
        Set of selected tracks to be kept.

    Returns
    -------
    pd.DataFrame
        Filtered and stitched tracks dataframe.
    """
    selected_track_ids = selected_track_ids.copy()

    start_df = []
    end_df = []
    track_ids = []
    for i, group in df.groupby("track_id", as_index=True):
        start_df.append(group.loc[group["t"].idxmin()])
        end_df.append(group.loc[group["t"].idxmax()])
        track_ids.append(int(i))

    start_df = pd.DataFrame(start_df, index=track_ids)
    # removing samples from divisions
    start_df = start_df[start_df["parent_track_id"] == NO_PARENT]

    end_df = pd.DataFrame(end_df, index=track_ids)
    # removing samples from divisions
    end_df = end_df[np.logical_not(end_df.index.isin(df["parent_track_id"]))]

    start_by_t = start_df.groupby("t")
    track_id_mapping = {i: i for i in track_ids}
    track_id_mapping[NO_PARENT] = NO_PARENT

    max_distance = estimate_drift(df)

    for t, end_group in end_df.groupby("t"):
        end_group = end_group[end_group.index.isin(selected_track_ids)]

        LOG.info(f"# {len(end_group)} track ends at t = {t}")
        if len(end_group) == 0:
            continue

        try:
            start_group = start_by_t.get_group(t + 1)
        except KeyError:
            continue

        start_group = start_group[
            np.logical_not(start_group.index.isin(selected_track_ids))
        ]

        LOG.info(f"# {len(end_group)} track ends at t = {t + 1}")
        if len(start_group) == 0:
            continue

        kdtree = KDTree(start_group[["z", "y", "x"]])
        dist, neighbors = kdtree.query(
            end_group[["z", "y", "x"]],
            distance_upper_bound=max_distance,
        )

        # removing invalid neighbors
        valid = dist != np.inf
        neighbors = neighbors[valid]
        neigh_df = pd.DataFrame(
            {"dist": dist[valid], "neighbor": start_group.index.values[neighbors]},
            index=end_group.index.values[valid],
        )

        neigh_df = neigh_df.sort_values(by=["dist"])
        neigh_df = neigh_df.drop_duplicates("neighbor", keep="first")

        LOG.info(
            f"stitching {neigh_df.index.values} to {neigh_df['neighbor'].values} at t = {t}"
        )

        for prev_id, new_id in zip(neigh_df["neighbor"], neigh_df.index):
            track_id_mapping[int(prev_id)] = int(new_id)
            selected_track_ids.update(get_subtree(graph, prev_id))

    LOG.info(f"track_id remapping {track_id_mapping}")

    df["track_id"] = df["track_id"].map(track_id_mapping)
    df["parent_track_id"] = df["parent_track_id"].map(track_id_mapping)
    df = df[df["track_id"].isin(selected_track_ids)]

    return df


def select_tracks_from_first_frame(
    data_config: DataConfig,
    first_frame: pd.DataFrame,
    df: pd.DataFrame,
    stitch_tracks: bool,
) -> None:
    """
    Selects only the subset of tracks rooted at the detection masks of `first_frame` array.

    Parameters
    ----------
    data_config : DataConfig
        Data configuration parameters.
    first_frame : pd.DataFrame
        Detection mask of first frame.
    df : pd.DataFrame
        Tracks dataframe.
    stitch_tracks : bool
        Stitches (connects) incomplete tracks nearby tracks on subsequent time point.
    """

    # query starting nodes data
    engine = sqla.create_engine(data_config.database_path)
    with Session(engine) as session:
        query = session.query(NodeDB.pickle).where(NodeDB.t == 0, NodeDB.selected)
        starting_nodes = [n for n, in query]

    root_centroids = np.asarray([n.centroid for n in starting_nodes])
    marker_centroids = np.asarray(
        [n.centroid for n in regionprops(first_frame, cache=False)]
    )
    D = cdist(marker_centroids, root_centroids)

    _, root_ids = linear_sum_assignment(D)

    selected_track_ids = set()
    graph = tracks_df_forest(df)

    for root in tqdm(root_ids, "Selecting tracks from first trame"):
        track_id = df.loc[starting_nodes[root].id, "track_id"]
        selected_track_ids.update(get_subtree(graph, track_id))

    if stitch_tracks:
        selected_df = stitch_tracks_df(graph, df, selected_track_ids)
    else:
        selected_df = df[df["track_id"].isin(selected_track_ids)].copy()

    return selected_df


def _validate_masks_path(output_dir: Path, overwrite: bool) -> None:
    """Validates existance of output masks paths."""
    mask_paths = list(output_dir.glob("mask*.tif"))
    if len(mask_paths) > 0:
        if overwrite:
            for path in mask_paths:
                path.unlink()
        else:
            raise ValueError(
                f"{output_dir}'s segmentation masks on already exists. Set `--overwrite` option to overwrite it."
            )


@curry
def _write_tiff_buffer(
    t: int,
    buffer: np.ndarray,
    output_dir: Path,
    scale: Optional[ArrayLike] = None,
    dilation_iters: int = 0,
) -> None:
    """Writes a single tiff stack into `output_dir` / "mask%03d.tif"

    Parameters
    ----------
    t : int
        Time index.
    buffer : np.ndarray
        Segmentation mask uint16 buffer.
    output_dir : Path
        Output directory.
    scale : Optional[ArrayLike], optional
        Mask rescaling factor, by default None
    dilation_iters: int
        Iterations of radius 1 morphological dilations on labels, applied after scaling, by default 0.
    """
    if scale is not None:
        buffer = zoom(
            buffer, scale[-buffer.ndim :], order=0, grid_mode=True, mode="grid-constant"
        )

    footprint = generate_binary_structure(buffer.ndim, 1)
    for _ in range(dilation_iters):
        dilated = grey_dilation(buffer, footprint=footprint)
        np.putmask(buffer, buffer == 0, dilated)

    # reducing mask sizes
    buffer = buffer.astype(
        np.promote_types(np.min_scalar_type(buffer.max()), np.uint16)
    )

    imwrite(output_dir / f"mask{t:03}.tif", buffer, compression="LZW")


def to_ctc(
    output_dir: Path,
    config: MainConfig,
    margin: int = 0,
    scale: Optional[Tuple[float]] = None,
    first_frame: Optional[ArrayLike] = None,
    dilation_iters: int = 0,
    stitch_tracks: bool = False,
    overwrite: bool = False,
) -> None:
    """
    Exports tracking results to cell-tracking challenge (http://celltrackingchallenge.net) format.

    Parameters
    ----------
    output_dir : Path
        Output directory to save segmentation masks and lineage graph
    config : DataConfig
        Configuration parameters.
    scale : Optional[Tuple[float]], optional
        Optional scaling of output segmentation masks, by default None
    margin : int
        Margin used to filter out nodes and splitting their tracklets
    first_frame : Optional[ArrayLike], optional
        Optional first frame detection mask to select a subset of tracks (e.g. Fluo-N3DL-DRO), by default None
    dilation_iters: int
        Iterations of radius 1 morphological dilations on labels, applied after scaling, by default 0.
    stitch_tracks: bool, optional
        Stitches (connects) incomplete tracks nearby tracks on subsequent time point, by default False
    overwrite : bool, optional
        Flag to overwrite existing `output_dir` content, by default False
    """
    if stitch_tracks and first_frame is None:
        raise NotImplementedError(
            "Tracks stitching only implemented for when `first_frame` is supplied."
        )

    output_dir.mkdir(exist_ok=True, parents=True)
    tracks_path = output_dir / "res_track.txt"

    _validate_masks_path(output_dir, overwrite)
    validate_and_overwrite_path(tracks_path, overwrite, "cli")

    df = solution_dataframe_from_sql(config.data_config.database_path)

    if len(df) == 0:
        raise ValueError("Solution is empty.")

    condition = None
    if scale is not None and not np.all(np.isclose(scale, 1.0)):
        condition = rescale_size_condition(scale)

    if margin > 0:
        if condition is None:
            condition = margin_filter_condition(config.data_config, margin)
        else:
            # mergin two conditions into one to avoid looping querying the db twice
            first_cond = condition
            second_cond = margin_filter_condition(config.data_config, margin)

            def condition(node: Node) -> bool:
                return first_cond(node) or second_cond(node)

    if condition is not None:
        df = filter_nodes_generic(config.data_config, df, condition)

    if len(df) == 0:
        raise ValueError("Solution is empty after filtering.")

    df = add_track_ids_to_tracks_df(df)

    if first_frame is not None:
        if scale is not None:
            first_frame = zoom(
                first_frame,
                1 / np.asarray(scale)[-first_frame.ndim :],
                order=0,
                grid_mode=True,
                mode="grid-constant",
            )

        df = select_tracks_from_first_frame(
            config.data_config,
            first_frame,
            df,
            stitch_tracks=stitch_tracks,
        )

    df["track_id"], fw, _ = relabel_sequential(df["track_id"].values)
    mask = df["parent_track_id"] != NO_PARENT
    df.loc[mask, "parent_track_id"] = fw[df.loc[mask, "parent_track_id"].values]

    # convert to CTC format and write output
    tracks_df = ctc_compress_forest(df)
    tracks_df.to_csv(tracks_path, sep=" ", header=False, index=False)

    LOG.info(f"CTC tracking data:\n{tracks_df}")

    export_segmentation_generic(
        config.data_config,
        df,
        _write_tiff_buffer(
            output_dir=output_dir, scale=scale, dilation_iters=dilation_iters
        ),
    )


def margin_filter_condition(
    data_config: DataConfig,
    margin: int,
) -> Callable[[Node], bool]:
    """Condition to check if nodes are completely inside margin.

    Parameters
    ----------
    data_config : DataConfig
        Data configuration parameters.
    margin : int
        yx-axis margin.

    Returns
    -------
    Callable[[Node], bool]
        Condition checking if node is completely inside the margins.
    """
    # ignoring time
    upper_lim = np.asarray(data_config.metadata["shape"][1:])
    upper_lim[-2:] -= margin  # only for y, x coordinates

    lower_lim = np.zeros_like(upper_lim)
    lower_lim[-2:] += margin  # only for y, x coordinates

    limits = np.concatenate((lower_lim, upper_lim))

    LOG.info(f"Using limits {limits} from {margin}")

    def _condition(node: Node) -> bool:
        return not intersects(node.bbox, limits)

    return _condition


def rescale_size_condition(
    scale: ArrayLike,
) -> Callable[[Node], bool]:
    """Condition to check if a node will disappear after scaling.

    Parameters
    ----------
    scale : ArrayLike
        Scaling factors.

    Returns
    -------
    Callable[[Node], bool]
        Condition to check if node is valid after scaling.
    """
    scale = np.asarray(scale)

    def _condition(node: Node) -> bool:
        ndim = node.mask.ndim
        return np.any(node.mask.shape * scale[-ndim:] < 1.0)

    return _condition



import logging
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike
from scipy.spatial.distance import pdist
from tqdm import tqdm

from ultrack.tracks.graph import (
    inv_tracks_df_forest,
    left_first_search,
    split_trees,
    tracks_df_forest,
)
from ultrack.utils.constants import NO_PARENT

LOG = logging.getLogger(__name__)


def _invert_forest(
    track_ids: np.ndarray,
    digraph: Dict[int, int],
) -> Dict[int, List[int]]:
    """
    Converts the directed graph (child -> parent) to a rooted forest (parent -> children).

    Parameters
    ----------
    track_ids : np.ndarray
        Array of track ids.
    digraph : Dict[int, int]
        Child -> parent directed graph (forest).

    Returns
    -------
    Dict[int, List[int]]
        Rooted forest (parent -> children)
    """
    forest = {}

    for id in track_ids:
        forest[id] = []

    for child, parent in digraph.items():
        forest[parent].append(child)

    return forest


def _accumulate_length(
    track_id: int,
    track_length: Dict[int, int],
    forest: Dict[int, List[int]],
) -> Tuple[int, List[int]]:
    """
    Accumulates the length of each root. For each branch the average length is considered.

    Parameters
    ----------
    track_id : int
        Current track id.
    track_length : Dict[int, int]
        Dictionary of tracklet lengths.
    forest : Dict[int, List[int]]
        Rooted forest.

    Returns
    -------
    Tuple[int, List[int]]
        Current subtree length and subtree ids.
    """
    subtree = [track_id]
    length = track_length[track_id]
    children = forest[track_id]

    if not children:
        return length, subtree

    branch_length = 0
    for child in children:
        child_length, child_subtree = _accumulate_length(child, track_length, forest)
        branch_length += child_length
        subtree += child_subtree

    length += branch_length / len(children) + 1

    return length, subtree


def sort_trees_by_length(
    df: Union[ArrayLike, pd.DataFrame],
    graph: Optional[Dict[int, int]] = None,
) -> List[pd.DataFrame]:
    """Sorts trees from the track graph by length (deepest tree path).

    Parameters
    ----------
    df : pd.DataFrame
        tracks dataframe.
    graph : Dict[int, int], optional
        Child -> parent tracks graph.
        Optional, if not provided it will be computed from the dataframe,
        must have `track_id` and `parent_track_id` columns.

    Returns
    -------
    List[pd.DataFrame]
        Sorted list of tracks dataframe.
    """
    if graph is None:
        graph = inv_tracks_df_forest(df)

    if not isinstance(df, pd.DataFrame):
        df = pd.DataFrame(df)
        df.rename(columns={0: "track_id"}, inplace=True)

    elif graph is None:
        graph = inv_tracks_df_forest(df)

    groups = df.groupby("track_id", as_index=False)

    length = groups.size()
    length.index = length["track_id"]
    track_ids = length["track_id"].to_numpy()
    length = length["size"].to_dict()

    forest = _invert_forest(track_ids, graph)

    roots = track_ids[np.isin(track_ids, list(graph.keys()), invert=True)]
    lengths = []
    subtrees = []

    for root in roots:
        tree_len, ids = _accumulate_length(root, length, forest)
        lengths.append(tree_len)
        subtrees.append(ids)

    sorted_trees = sorted(zip(lengths, subtrees), reverse=True)
    return [
        pd.concat([groups.get_group(i) for i in subtree]) for _, subtree in sorted_trees
    ]


def sort_track_ids(
    tracks_df: pd.DataFrame,
) -> np.ndarray:
    """
    Sort track IDs in a given DataFrame representing tracks in a way that maintains the left-first
    order of the binary tree formed by their parent-child relationships.

    Parameters
    ----------
    tracks_df : pd.DataFrame
        A DataFrame containing information about tracks, where each row represents a track and
        contains at least two columns - "track_id" and "track_parent_id". The "track_id" column
        holds unique track IDs, and the "track_parent_id" column contains the parent track IDs
        for each track. The DataFrame should have a consistent parent-child relationship, forming
        one or multiple binary trees.

    Returns
    -------
    np.ndarray
        A NumPy array containing the sorted track IDs based on the left-first traversal of the
        binary trees formed by the parent-child relationships.

    Example
    -------
    >>> import pandas as pd
    >>> import numpy as np
    >>> data = {
    ...     "track_id": [1, 2, 3, 4, 5, 6, 7],
    ...     "track_parent_id": [None, 1, 1, 2, 2, 3, 3],
    ... }
    >>> tracks_df = pd.DataFrame(data)
    >>> sorted_track_ids = sort_track_ids(tracks_df)
    >>> print(sorted_track_ids)
    [4 2 5 1 6 3 7]
    """
    graph = tracks_df_forest(tracks_df, numba_dict=True)
    roots = graph.pop(NO_PARENT)

    sorted_track_ids = []
    for root in roots:
        sorted_track_ids += left_first_search(root, graph)

    return np.asarray(sorted_track_ids)


def sort_trees_by_max_radius(
    df: Union[ArrayLike, pd.DataFrame],
    scale: Optional[ArrayLike] = None,
    metric: Union[Callable, str] = "euclidean",
) -> List[pd.DataFrame]:
    """Sorts trees from the track graph by radius (distance between nodes at the same time point).

    Parameters
    ----------
    df : pd.DataFrame
        tracks dataframe.
    scale : ArrayLike, optional
        Spatial domain scale.
    metric : Union[Callable, str], optional
        Distance metric, see scipy.spatial.distance.pdist for more information.

    Returns
    -------
    List[pd.DataFrame]
        Sorted list of tracks dataframe.
    """
    if not isinstance(df, pd.DataFrame):
        if df.shape[1] == 4:
            cols = ["track_id", "t", "y", "x"]
        elif df.shape[1] == 5:
            cols = ["track_id", "t", "z", "y", "x"]
        else:
            raise ValueError("Only 2D and 3D data is supported")
        df = pd.DataFrame(df, columns=cols)

    spatial_cols = [c for c in ("z", "y", "x") if c in df.columns]

    if scale is not None:
        df = df.copy()
        df[spatial_cols] *= scale

    LOG.info("Splitting trees")
    trees = split_trees(df)
    max_radii = []

    def _max_dist(df: pd.DataFrame) -> float:
        if df.shape[0] == 1:
            return 0
        return pdist(df, metric=metric).max()

    LOG.info("Computing max radii per tree")
    for tree in tqdm(trees, "Computing max radius per tree"):
        max_radii.append(tree.groupby("t")[spatial_cols].apply(_max_dist).values.max())

    LOG.info("Sorting trees")
    sorted_trees = sorted(zip(max_radii, trees), reverse=True, key=lambda x: x[0])
    sorted_trees = [tree for _, tree in sorted_trees]

    if scale is not None:
        for tree in sorted_trees:
            tree.loc[:, spatial_cols] /= scale

    return sorted_trees



import enum
import logging
from abc import abstractmethod
from typing import List, Optional

import napari
from magicgui.widgets import ComboBox, PushButton
from napari.layers import Image, Labels

from ultrack.core.database import NodeDB, is_table_empty
from ultrack.core.segmentation.node import Node
from ultrack.widgets._generic_data_widget import GenericDataWidget

logging.basicConfig()
logging.getLogger("sqlachemy.engine").setLevel(logging.INFO)

LOG = logging.getLogger(__name__)


class GenericAnnotationWidget(GenericDataWidget):
    def __init__(
        self,
        viewer: napari.Viewer,
        label: str,
        annot_opts: enum.IntEnum,
        suffix: str,
        sample_size: int = 50,
    ) -> None:

        # before init due to config initialization
        self._nodes: List[Node] = []
        self._sample_size = sample_size
        self._suffix = suffix
        self._mask_layer_name = f"Current Node Mask ~ {self._suffix}"

        self._next_btn = PushButton(text="Next")
        self._prev_btn = PushButton(text="Prev")
        self._confirm_btn = PushButton(text="Confirm", enabled=False)

        super().__init__(viewer, label)

        self._next_btn.changed.connect(self._on_next)
        self.append(self._next_btn)

        self._prev_btn.changed.connect(self._on_prev)
        self.append(self._prev_btn)

        self._annot_w = ComboBox(label="Annot.", choices=annot_opts)
        self.append(self._annot_w)

        self._confirm_btn.changed.connect(self._on_confirm)
        self.append(self._confirm_btn)

        self.list_index = 0

    def _on_config_changed(self) -> None:
        self._nodes: List[Node] = []

        for layer in list(self._viewer.layers):
            if layer.name.endswith(self._suffix):
                self._viewer.layers.remove(layer.name)

        # must be after cleanup
        self.list_index = 0

    @abstractmethod
    def _query_samples(self) -> List[Node]:
        pass

    def _is_valid_index_at_list_end(self, index: int) -> bool:
        if index >= len(self._nodes) and not is_table_empty(self.config, NodeDB):
            self._nodes += self._query_samples()
        return index < len(self._nodes) - 1

    @property
    def list_index(self) -> int:
        return self._list_index

    @list_index.setter
    def list_index(self, value: int) -> int:
        self._list_index = value
        self._prev_btn.enabled = value > 0
        self._next_btn.enabled = self._is_valid_index_at_list_end(value)
        self._confirm_btn.enabled = len(self._nodes) > 0
        if 0 <= self._list_index < len(self._nodes):
            self._load_node(self._nodes[self._list_index])

    def _on_next(self) -> None:
        self.list_index += 1

    def _on_prev(self) -> None:
        self.list_index -= 1

    def _load_node(self, node: Node) -> None:

        shape = tuple(self.config.metadata["shape"])
        for layer in list(self._viewer.layers):
            if (
                isinstance(layer, (Image, Labels))
                and layer.data.shape == shape
                and not layer.name.endswith(self._suffix)
            ):

                data, state, type_str = layer.as_layer_data_tuple()
                name = state["name"]
                crop_name = name + self._suffix
                if crop_name in self._viewer.layers:
                    state["visible"] = self._viewer.layers[crop_name].visible
                    self._viewer.layers.remove(crop_name)

                data = node.roi(data[node.time])

                state["translate"] = (
                    state["translate"][-data.ndim :] + node.bbox[: data.ndim]
                )
                state["scale"] = state["scale"][-data.ndim :]
                state["name"] = crop_name

                # removing this matrices for now
                state.pop("rotate")
                state.pop("shear")
                state.pop("affine")

                # was causing issues between 2d and 3d
                state.pop("interpolation", None)

                self._viewer._add_layer_from_data(data, state, type_str)

        try:
            layer = self._viewer.layers[self._mask_layer_name]
            layer.data = node.mask
            layer.translate = node.bbox[: node.mask.ndim]
        except KeyError:
            layer = self._viewer.add_labels(
                node.mask,
                name=self._mask_layer_name,
                translate=node.bbox[: node.mask.ndim],
            )
            layer.bind_key("Enter", self._on_confirm)

        self._annot_w.value = self.get_annotation(node.id)
        self._viewer.layers.move(self._viewer.layers.index(self._mask_layer_name), -1)

        if self._viewer.dims.ndim == node.mask.ndim:
            self._viewer.dims.set_point(range(len(node.centroid)), node.centroid)
        else:
            self._viewer.dims.set_point(
                range(len(node.centroid) + 1), (node.time, *node.centroid)
            )

        self._viewer.camera.center = node.centroid

    def _on_confirm(self, layer: Optional[Labels] = None) -> None:
        if not self._confirm_btn.enabled:
            # required by key binding
            return

        # layer required by napari bind_key
        self.set_annotation(self._nodes[self.list_index].id, self._annot_w.value)

        if self._next_btn.enabled:
            self._next_btn.clicked.emit()

    @abstractmethod
    def get_annotation(self, index: int) -> enum.IntEnum:
        pass

    @abstractmethod
    def set_annotation(self, index: int, annot: enum.IntEnum) -> None:
        pass



from typing import Tuple

import higra as hg
import numpy as np
import pytest
from skimage.segmentation import find_boundaries

from ultrack.core.segmentation.hierarchy import create_hierarchies
from ultrack.core.segmentation.vendored.hierarchy import to_labels


@pytest.mark.parametrize(
    "segmentation_mock_data",
    [{"size": 128, "n_dim": 2}, {"size": 64, "n_dim": 3}],
    indirect=True,
)
def test_horizontal_cut(
    segmentation_mock_data: Tuple[np.ndarray, np.ndarray, np.ndarray],
) -> None:
    foreground, edge, _ = segmentation_mock_data
    hierarchies = create_hierarchies(
        foreground,
        edge,
        hierarchy_fun=hg.watershed_hierarchy_by_dynamics,
        min_area=0,
        min_area_factor=1,
        cut_threshold=0.1,
    )

    segmentation = to_labels(hierarchies, edge.shape)
    contours = find_boundaries(segmentation)

    # differente between contours
    difference = np.logical_xor(contours, edge).sum()

    # arbitrary proportion to take into account watershed tie-zoes
    assert difference < edge.size * 0.005



from pathlib import Path
from typing import Callable, Tuple

import napari
import pytest
import zarr
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from ultrack.config import MainConfig
from ultrack.core.database import NodeDB, NodeSegmAnnotation
from ultrack.widgets import NodeAnnotationWidget

pytest.skip(reason="DEPRECATED", allow_module_level=True)


def test_node_annotation_widget(
    make_napari_viewer: Callable[[], napari.Viewer],
    tracked_database_mock_data: MainConfig,
    timelapse_mock_data: Tuple[zarr.Array, zarr.Array, zarr.Array],
    tmp_path: Path,
    request,
) -> None:
    # NOTE: Use "--show-napari-viewer" to show viewer, useful when debugging

    config = tracked_database_mock_data
    config.data_config.working_dir = tmp_path

    viewer = make_napari_viewer()

    for arr in timelapse_mock_data:
        viewer.add_image(arr)

    widget = NodeAnnotationWidget(viewer)

    viewer.window.add_dock_widget(widget)

    widget.config = config.data_config
    # there was a bug where _nodes would be empty after setting up the config
    assert len(widget._nodes) > 0

    if request.config.getoption("--show-napari-viewer"):
        napari.run()
        return

    assert widget._mask_layer_name in viewer.layers
    assert widget._next_btn.enabled
    assert not widget._prev_btn.enabled
    assert widget.list_index == 0

    widget._next_btn.clicked.emit()
    assert widget._prev_btn.enabled
    assert widget.list_index == 1

    widget._prev_btn.clicked.emit()
    assert not widget._prev_btn.enabled
    assert widget.list_index == 0

    widget._annot_w.value = NodeSegmAnnotation.OVERSEGMENTED
    widget._confirm_btn.clicked.emit()
    assert widget.list_index == 1

    engine = create_engine(widget.config.database_path)
    with Session(engine) as session:
        table_length = session.query(NodeDB.id).count()

    while widget._next_btn.enabled:
        # a bit slow, but worth it
        widget._next_btn.clicked.emit()

    assert widget.list_index == table_length - 1

    widget._annot_w.value = NodeSegmAnnotation.CORRECT
    widget._confirm_btn.clicked.emit()

    # checking if index won't advance when confirming the last node
    assert widget.list_index == table_length - 1

    with Session(engine) as session:
        query = session.query(NodeDB.id)
        n_correct = query.where(
            NodeDB.segm_annotation == NodeSegmAnnotation.CORRECT
        ).count()
        n_oversegmented = query.where(
            NodeDB.segm_annotation == NodeSegmAnnotation.OVERSEGMENTED
        ).count()
        n_undersegmented = query.where(
            NodeDB.segm_annotation == NodeSegmAnnotation.UNDERSEGMENTED
        ).count()

    assert n_correct == 1
    assert n_oversegmented == 1
    assert n_undersegmented == 0



import click
from rich import print


@click.command("check_gurobi")
def check_gurobi_cli() -> None:

    try:
        import gurobipy as gp

        print("Gurobi module imported successfully.")

        # Attempt to create a Gurobi environment. This will check the license.
        gp.Env()
        print("Gurobi environment created successfully.")

    except ImportError:
        print(
            "Gurobi module not found. Ensure it is installed and the Python path is set correctly."
        )
        print("Using `conda install gurobi -c gurobi` is recommended.")
    except gp.GurobiError as e:
        print(f"GurobiError encountered: {e.message}")
        print("There may be a problem with your Gurobi license or setup.")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")



from pathlib import Path
from typing import Callable, Dict

import napari
import numpy as np
import pytest
import sqlalchemy as sqla
from sqlalchemy.orm import Session

from ultrack.config import MainConfig
from ultrack.core.database import NodeDB
from ultrack.core.segmentation.node import Node
from ultrack.widgets import HypothesesVizWidget

pytest.skip(reason="DEPRECATED", allow_module_level=True)


def _is_sorted(nodes: Dict[int, Node]) -> bool:
    nodes = list(nodes.values())
    return all(nodes[i].area <= nodes[i + 1].area for i in range(len(nodes) - 1))


def test_hypotheses_viz_widget(
    make_napari_viewer: Callable[[], napari.Viewer],
    linked_database_mock_data: MainConfig,
    tmp_path: Path,
    request,
) -> None:
    # NOTE: Use "--show-napari-viewer" to show viewer, useful when debugging

    config = linked_database_mock_data
    config.data_config.working_dir = tmp_path

    viewer = make_napari_viewer()

    widget = HypothesesVizWidget(viewer)
    widget.config = config.data_config
    viewer.window.add_dock_widget(widget)

    with pytest.warns(UserWarning):
        assert widget._time == 0

    with pytest.warns(UserWarning):
        widget._load_btn.clicked.emit()

    assert _is_sorted(widget._nodes)

    # forcing every node to be drawn
    widget._area_threshold_w.value = widget._area_threshold_w.max

    labels = viewer.layers[widget._segm_layer_name].data

    engine = sqla.create_engine(config.data_config.database_path)
    with Session(engine) as session:
        query = session.query(NodeDB.pickle).where(NodeDB.t == 0)

    # checking if every node was drawn
    for (node,) in query:
        assert np.all(labels[node.mask_indices()])

    widget._load_neighbors(0)
    assert widget._link_layer_name not in viewer.layers

    # testing using last `node`
    widget._load_neighbors(node.id)
    assert widget._link_layer_name in viewer.layers

    if request.config.getoption("--show-napari-viewer"):
        napari.run()



"""
Auxiliary file to log and debug main widget usage.
"""

import logging

logging.basicConfig(filename="ultrackwidget.log", filemode="w", level=logging.INFO)

import napari

from ultrack.widgets.ultrackwidget import UltrackWidget

if __name__ == "__main__":
    viewer = napari.Viewer()
    widget = UltrackWidget(viewer)
    viewer.window.add_dock_widget(widget)
    napari.run()



import asyncio
import io
import json
import logging
import threading
import time
import traceback as tb
from contextlib import redirect_stderr, redirect_stdout
from typing import Optional

from fastapi import HTTPException, WebSocket
from starlette.websockets import WebSocketState

from ultrack.api.database import Experiment, ExperimentStatus

LOG = logging.getLogger(__name__)


def _mimic_carriage_return(text: str) -> str:
    """Mimic the carriage return `\r` behavior of the output text. Otherwise, text
    outputs that heavily relies on updating the line with the carriage return,
    like tqdm for example, will flood the websocket with multiple lines.
    """
    parts = text.split("\r")
    current_line = ""
    final_output = []

    for part in parts:
        if "\n" in part:
            lines = part.split("\n")
            current_line += lines[0]
            final_output.append(current_line)
            final_output.extend(lines[1:-1])
            current_line = lines[-1]
        else:
            current_line = part

    if current_line:
        final_output.append(current_line)

    return "\n".join(final_output)


class ProcessedException(HTTPException):
    """Exception that has been processed by the API."""


async def raise_error(
    e: Exception,
    exp: Optional[Experiment] = None,
    ws: Optional[WebSocket] = None,
    additional_info: Optional[str] = None,
) -> None:
    """Raise an error and send it to the websocket.

    Finishes the experiment and sends the error message to the websocket.  If the
    exception is a ProcessedException, it will not be raised again.

    Parameters
    ----------
    e : Exception
        The exception to be raised.
    exp : Optional[Experiment]
        The experiment that failed. Defaults to None if the exception is not related to
        an initialized experiment.
    ws : Optional[WebSocket]
        The websocket to send the error message. Defaults to None if the exception is
        not related to a connected websocket.
    additional_info : Optional[str]
        Additional information to be sent to the websocket. Defaults to None.

    Raises
    ------
    ProcessedException
        Wraps the original exception to indicate that it has been processed by the API
        and should not be raised again.
    """
    if not isinstance(e, ProcessedException):
        from ultrack.api.app import finish_experiment

        LOG.error(e)
        msg = ""
        if exp is not None:
            msg = f"Experiment {exp.id} failed.\n"
            exp.status = ExperimentStatus.ERROR
            await finish_experiment(ws, exp)
        if additional_info is not None:
            msg += additional_info + "\n"
        msg += f"Exception: {str(e)}\n"
        msg += "Traceback:\n" + tb.format_exc()
        if ws is not None and ws.client_state == WebSocketState.CONNECTED:
            await ws.send_text(msg)
        raise ProcessedException(status_code=500, detail=str(msg))


class UltrackWebsocketLogger:
    """Context manager to log experiment output to a websocket.

    This context manager logs the stdout and stderr of a block of code to a websocket.

    Attributes
    ----------
    websocket : WebSocket
        The websocket to send the logs.
    experiment : Experiment
        The experiment to be logged.
    ctx_stdout_switcher : redirect_stdout
        The context manager wrapped to redirect the stdout.
    ctx_stderr_switcher : redirect_stderr
        The context manager wrapped to redirect the stderr.
    stdout : io.StringIO
        The buffer to store the stdout.
    stderr : io.StringIO
        The buffer to store the stderr.
    stop_logger_event : threading.Event
        The event to stop the logger thread.

    Example
    -------
    >>> experiment = ...
    >>> websocket = ...
    >>> async with UltrackWebsocketLogger(websocket, experiment):
    ...     await start_experiment(websocket, experiment)
    ...     ...
    ...     await segment_link_and_solve(websocket, experiment)
    ...     ...
    ... await finish_experiment(websocket, experiment)
    """

    interruption_handlers: dict[int, threading.Event] = {}

    @staticmethod
    def register_interruption_handler(exp_id: int) -> threading.Event:
        """Register an interruption thread handler for the given experiment id.

        It should be used to interrupt the experiment execution thread
        from another thread.

        Parameters
        ----------
        exp_id : int
            The experiment id.

        Returns
        -------
        threading.Event
            The interruption handler.

        See Also
        --------
        UltrackWebsocketLogger.request_interruption : request the interruption of the
            experiment execution thread.
        """
        handler = threading.Event()
        UltrackWebsocketLogger.interruption_handlers[exp_id] = handler
        return handler

    @staticmethod
    def request_interruption(exp_id: int) -> None:
        """Request the interruption of the experiment execution thread.

        Parameters
        ----------
        exp_id : int
            The experiment id.

        See Also
        --------
        UltrackWebsocketLogger.register_interruption_handler : register an interruption
            handler for the given experiment id.
        """
        # handler = UltrackWebsocketLogger.interruption_handlers.get(exp_id)
        # if handler is not None:
        #     handler.set()
        # del UltrackWebsocketLogger.interruption_handlers[exp_id]
        raise NotImplementedError(
            "Interruption handler is not implemented yet due to asyncio limitations."
        )

    def __init__(
        self,
        websocket: WebSocket,
        experiment: Experiment,
        tick_rate: float = 1.0,
    ):
        """Initialize the UltrackWebsocketLogger context manager.

        Parameters
        ----------
        websocket : WebSocket
            The websocket to send the logs.
        experiment : Experiment
            The experiment to be logged.
        tick_rate : float
            The rate (in sec.) to send the logs to the websocket, by default 5.0 (sec.).
        """
        self.websocket = websocket
        self.experiment = experiment
        self.ctx_stdout_switcher = redirect_stdout(io.StringIO())
        self.ctx_stderr_switcher = redirect_stderr(io.StringIO())
        self.stdout = None
        self.stderr = None

        async def run_thread():
            """Logging thread function to be run in the background."""
            if self.stdout is not None:
                experiment.std_log = _mimic_carriage_return(self.stdout.getvalue())
            if self.stderr is not None:
                experiment.err_log = _mimic_carriage_return(self.stderr.getvalue())
            if websocket.state == WebSocketState.DISCONNECTED:
                return False
            await websocket.send_json(json.loads(experiment.model_dump_json()))
            return True

        def between_callback(stop_logger_event: threading.Event):
            """Helper function to run the logging thread in the background. It is
            necessary to have it since the asyncio loop should run in a separate thread.
            """
            while not stop_logger_event.is_set():
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                still_connected = loop.run_until_complete(run_thread())
                if not still_connected:
                    stop_logger_event.set()
                    raise Exception(
                        "Websocket was disconnected. "
                        "The experiment will be interrupted."
                    )
                loop.close()
                if (
                    experiment.id in UltrackWebsocketLogger.interruption_handlers
                    and UltrackWebsocketLogger.interruption_handlers[
                        experiment.id
                    ].is_set()
                ):
                    stop_logger_event.set()
                    raise Exception("Experiment was interrupted.")
                time.sleep(tick_rate)

        self.stop_logger_event = threading.Event()

        self.thread = threading.Thread(
            target=between_callback,
            args=(self.stop_logger_event,),
        )

    async def __aenter__(self) -> "UltrackWebsocketLogger":
        """Enter the context manager.

        It starts the logging thread and redirects the stdout and stderr to the
        buffers.
        """
        self.stdout = self.ctx_stdout_switcher.__enter__()
        self.stderr = self.ctx_stderr_switcher.__enter__()
        self.thread.start()

        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> bool:
        """Exit the context manager.

        It stops the logging thread and redirects the stdout and stderr back to the
        original streams. It also sends the logs to the websocket if an exception was
        raised.
        """
        if exc_type is not None:
            self.stop_logger_event.set()
            self.thread.join()
            exc_obj = exc_type(exc_value).with_traceback(traceback)
            if not isinstance(exc_obj, ProcessedException):
                await raise_error(exc_obj, self.experiment, ws=self.websocket)
        self.experiment.std_log = _mimic_carriage_return(self.stdout.getvalue())
        self.experiment.err_log = _mimic_carriage_return(self.stderr.getvalue())
        self.stop_logger_event.set()
        self.thread.join()
        self.ctx_stdout_switcher.__exit__(exc_type, exc_value, traceback)
        self.ctx_stderr_switcher.__exit__(exc_type, exc_value, traceback)
        return False



import logging
from typing import Optional, Protocol, Union

import numpy as np
import pandas as pd
import sqlalchemy as sqla
from numpy.typing import ArrayLike
from rich.logging import RichHandler
from sqlalchemy.orm import Session

from ultrack.config.config import MainConfig
from ultrack.core.database import LinkDB, set_node_values
from ultrack.core.linking.features import get_links_features
from ultrack.core.match_gt import UnmatchedNode, match_to_ground_truth
from ultrack.core.segmentation import get_nodes_features

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)
LOG.addHandler(RichHandler())


class ProbabilisticClassifier(Protocol):
    def fit(self, X: ArrayLike, y: ArrayLike) -> "ProbabilisticClassifier": ...

    def predict_proba(self, X: ArrayLike) -> ArrayLike: ...


def _validate_classifier(
    classifier: Optional[ProbabilisticClassifier],
) -> ProbabilisticClassifier:
    """
    Validates the classifier if it is not provided.
    """
    if classifier is None:
        try:
            from catboost import CatBoostClassifier
        except ImportError as e:
            raise ImportError(
                "`catboost` is required if classifier is not provided.\n"
                "Please install it with `pip install catboost` or `pip install 'ultrack[ml]'`."
            ) from e
        classifier = CatBoostClassifier(allow_const_label=True, silent=True)

    return classifier


def add_nodes_prob(
    config: MainConfig,
    indices: ArrayLike,
    probs: ArrayLike,
) -> None:
    """
    Add nodes' probabilities to the segmentation/tracking database.

    Parameters
    ----------
    config : MainConfig
        Main configuration parameters.
    indices : ArrayLike
        Nodes' indices database index.
    probs : ArrayLike
        Nodes' probabilities.
    """
    if probs.ndim == 2:
        probs = probs[:, -1]

    set_node_values(
        config.data_config,
        indices,
        node_prob=probs,
    )


def add_links_prob(
    config: MainConfig,
    indices: ArrayLike,
    probs: ArrayLike,
) -> None:
    """
    Add links' probabilities to the tracking database.

    NOTE: this only updates the weights of the existing links, it does not add new links.
    If you want to add new links, use `add_links` from `ultrack.core.linking.processing`.

    Parameters
    ----------
    config : MainConfig
        Main configuration parameters.
    indices : ArrayLike
        Nx2 array of (target_id, source_id) indices.
    probs : ArrayLike
        Links' probabilities.
    """
    if indices.ndim != 2:
        raise ValueError(f"Indices must be a 2D array, got {indices.shape}")

    if indices.shape[1] != 2:
        raise ValueError(f"Indices must have 2 columns, got {indices.shape[1]}")

    indices = np.asarray(indices, dtype=int)

    if probs.ndim == 2:
        probs = probs[:, -1]

    probs_df = pd.DataFrame(
        probs,
        index=pd.MultiIndex.from_arrays(
            [indices[:, 0], indices[:, 1]],
            names=["target_id", "source_id"],
        ),
        columns=["weight"],
    )

    LOG.info(
        "Adding %d links probabilities to the database",
        len(indices),
    )

    engine = sqla.create_engine(config.data_config.database_path)

    with Session(engine) as session:
        query = session.query(LinkDB.id, LinkDB.target_id, LinkDB.source_id)
        link_df = pd.read_sql(
            query.statement, session.bind, index_col=["target_id", "source_id"]
        )

        update = (
            sqla.update(LinkDB)
            .where(LinkDB.id == sqla.bindparam("link_id"))
            .values(weight=sqla.bindparam("weight"))
        )
        probs_df = probs_df.join(link_df, how="left")

        assert not probs_df.isna().any().any()

        session.connection().execute(
            update,
            [
                {"link_id": link_id, "weight": weight}
                for link_id, weight in zip(probs_df["id"], probs_df["weight"])
            ],
            execution_options={"synchronize_session": False},
        )
        session.commit()

    LOG.info("Successfully added %d links probabilities to the database", len(indices))


def predict_nodes_prob(
    config: MainConfig,
    classifier: ProbabilisticClassifier,
    insert_prob: bool = True,
    persistence_features: bool = False,
    coord_features: bool = False,
) -> pd.Series:
    """
    Predicts the probabilities of the nodes' features.

    Parameters
    ----------
    config : MainConfig
        Main configuration parameters.
    classifier : ProbabilisticClassifier
        Probabilistic classifier object.
    insert_prob : bool, optional
        Whether to insert the probabilities to the database, by default True.
    persistence_features : bool, optional
        Whether to include persistence features, by default False.
    coord_features : bool, optional
        Whether to include coordinate (t, (z), y, x) features, by default False.

    Returns
    -------
    pd.Series
        Nodes' probabilities.
    """
    features = get_nodes_features(config, include_persistence=persistence_features)

    if not coord_features:
        features = features.drop(columns=["t", "z", "y", "x"], errors="ignore")

    LOG.info("Predicting classifier with features: %s", str(features.columns))

    probs = classifier.predict_proba(features)

    if probs.ndim == 2:
        probs = probs[:, -1]

    if insert_prob:
        add_nodes_prob(config, features.index, probs)

    return pd.Series(probs, index=features.index)


def fit_nodes_prob(
    config: MainConfig,
    ground_truth: Union[ArrayLike, pd.Series],
    classifier: Optional[ProbabilisticClassifier] = None,
    remove_no_overlap: bool = True,
    insert_prob: bool = True,
    persistence_features: bool = False,
    coord_features: bool = False,
) -> ProbabilisticClassifier:
    """
    Fit a probabilistic classifier to the nodes' features.

    Parameters
    ----------
    config : MainConfig
        Main configuration parameters.
    ground_truth : Union[ArrayLike, pd.Series]
        Ground-truth labels, either a:
        * timelapse array of labels (T, (Z), Y, X) with the same shape as the input data.
        * pandas Series indexed by the nodes' indices.
    classifier : ProbabilisticClassifier
        Probabilistic classifier object.
        Classifier is fit in-place.
        If not provided, it will use `catboost.CatBoostClassifier`.
    remove_no_overlap : bool, optional
        Whether to remove **NO_OVERLAP** nodes (-1) from the ground-truth.
        Classification will compare **matched** (>0) vs **unmatched** nodes (0).
    insert_prob : bool, optional
        Whether to insert the probabilities to the database, by default True.
    persistence_features : bool, optional
        Whether to include persistence features, by default False.
    coord_features : bool, optional
        Whether to include coordinate (t, (z), y, x) features, by default False.

    Returns
    -------
    ProbabilisticClassifier
        Fitted probabilistic classifier.
    """
    features = get_nodes_features(config, include_persistence=persistence_features)

    if not coord_features:
        features = features.drop(columns=["t", "z", "y", "x"], errors="ignore")

    if "id" in features.columns:
        raise ValueError("Features cannot contain 'id' column")

    if not isinstance(ground_truth, (pd.Series, pd.DataFrame)):
        ground_truth = match_to_ground_truth(config, ground_truth)["gt_track_id"]

    # removing cells with no ground-truth, therefore we cannot decide they are blocked or not
    no_overlap = ground_truth != UnmatchedNode.NO_OVERLAP

    LOG.info("Found %f of nodes with no overlap", no_overlap.mean())

    if remove_no_overlap:
        ground_truth = ground_truth[no_overlap]

    # making it a binary classification problem
    ground_truth = (
        ground_truth > UnmatchedNode.BLOCKED
    )  # (includes NO_OVERLAP as negative class)

    LOG.info("Fitting classifier with features: %s", str(features.columns))

    classifier = _validate_classifier(classifier)

    if ground_truth.dtype != bool:
        raise ValueError(
            f"Ground-truth dataframe must be a binary classification problem (bool dtype), got {ground_truth.dtype}"
        )

    # Fit the classifier
    classifier.fit(features.loc[ground_truth.index], ground_truth)

    if insert_prob:
        LOG.info(
            "Predicting nodes probabilities with features: %s", str(features.columns)
        )

        probs = classifier.predict_proba(features)
        add_nodes_prob(config, features.index, probs)

    return classifier


def select_competing_links(
    links_df: pd.DataFrame,
    gt_link: pd.Series,
) -> tuple[pd.DataFrame, pd.Series]:
    """
    Selects the competing links from the ground-truth.

    Parameters
    ----------
    links_df : pd.DataFrame
        Links dataframe with features.
    gt_link : pd.Series
        Ground-truth (0/1)-links series.

    Returns
    -------
    tuple[pd.DataFrame, pd.Series]
        Links dataframe and ground-truth with the selected competing links.
    """
    links_df["gt_link"] = gt_link
    target_ids = links_df.index.get_level_values(0)

    has_competing_links = links_df.groupby(target_ids)["gt_link"].transform(
        lambda x: x.any()
    )
    LOG.info(
        "Found %d competing links out of %d candidates",
        has_competing_links.sum(),
        len(target_ids),
    )

    links_df = links_df[has_competing_links]

    return links_df.drop(columns=["gt_link"]), links_df["gt_link"]


def add_links_gt(
    links_df: pd.DataFrame,
    ground_truth: pd.Series,
) -> pd.Series:
    """
    Adds the ground-truth label to the links dataframe.

    Parameters
    ----------
    links_df : pd.DataFrame
        Links dataframe.
    ground_truth : pd.Series
        Ground-truth labels.

    Returns
    -------
    pd.Series:
        Series with the ground-truth label.
    """
    target_ids = links_df.index.get_level_values(0)
    source_ids = links_df.index.get_level_values(1)

    matched_nodes = ground_truth > UnmatchedNode.BLOCKED
    matched_nodes = (
        matched_nodes.loc[source_ids].to_numpy()
        & matched_nodes.loc[target_ids].to_numpy()
    )

    correct_links = (
        ground_truth.loc[source_ids].to_numpy()
        == ground_truth.loc[target_ids].to_numpy()
    )

    gt_links = pd.Series(
        correct_links & matched_nodes,
        index=pd.MultiIndex.from_arrays(
            [target_ids, source_ids], names=["target_id", "source_id"]
        ),
        name="gt_link",
    )

    LOG.info(
        "Found %d ground-truth links out of %d candidates",
        gt_links.sum(),
        len(gt_links),
    )

    return gt_links


def predict_links_prob(
    config: MainConfig,
    classifier: ProbabilisticClassifier,
    *,
    insert_prob: bool = True,
    persistence_features: bool = False,
    **kwargs,
) -> pd.Series:
    """
    Predicts the probabilities of the links' features.

    Parameters
    ----------
    config : MainConfig
        Main configuration parameters.
    classifier : ProbabilisticClassifier
        Probabilistic classifier object.
    insert_prob : bool, optional
        Whether to insert the probabilities to the database, by default True.
    persistence_features : bool, optional
        Whether to include persistence features, by default False.
    **kwargs : dict
        Keyword arguments passed to `get_nodes_features` to compute the features.
        The links' features are the difference between the source and target nodes' features.
    """
    features = get_links_features(
        config,
        include_persistence=persistence_features,
        **kwargs,
    )

    for col in ["target_id", "source_id"]:
        if col in features.columns:
            raise ValueError(f"Features cannot contain '{col}' column")

    LOG.info("Predicting links probabilities with features: %s", str(features.columns))

    probs = classifier.predict_proba(features)

    if probs.ndim == 2:
        probs = probs[:, -1]

    if insert_prob:
        add_links_prob(
            config,
            np.asarray(features.index.to_list()),
            probs,
        )

    return pd.Series(probs, index=features.index, name="link_prob")


def fit_links_prob(
    config: MainConfig,
    ground_truth: Union[ArrayLike, pd.Series],
    classifier: Optional[ProbabilisticClassifier] = None,
    remove_no_overlap: bool = True,
    insert_prob: bool = True,
    persistence_features: bool = False,
    **kwargs,
) -> ProbabilisticClassifier:
    """
    Fit a probabilistic classifier to the links' features.

    Parameters
    ----------
    config : MainConfig
        Main configuration parameters.
    ground_truth : Union[ArrayLike, pd.Series]
        Ground-truth labels, either a:
        * timelapse array of labels (T, (Z), Y, X) with the same shape as the input data.
        * pandas Series indexed by the nodes' indices.
    classifier : ProbabilisticClassifier
        Probabilistic classifier object.
        Classifier is fit in-place.
        If not provided, it will use `catboost.CatBoostClassifier`.
    remove_no_overlap : bool, optional
        When true it will only consider negative-labeled links when they are competing (overlap)
        with an existing positive-labeled link.
    insert_prob : bool, optional
        Whether to insert the probabilities to the database, by default True.
    persistence_features : bool, optional
        Whether to include persistence features, by default False.
    **kwargs : dict
        Keyword arguments passed to `get_nodes_features` to compute the features.
        The links' features are the difference between the source and target nodes' features.
    """
    features = get_links_features(
        config,
        include_persistence=persistence_features,
        **kwargs,
    )

    for col in ["target_id", "source_id"]:
        if col in features.columns:
            raise ValueError(f"Features cannot contain '{col}' column")

    if not isinstance(ground_truth, (pd.Series, pd.DataFrame)):
        ground_truth = match_to_ground_truth(config, ground_truth)["gt_track_id"]

    LOG.info("Ground-truth labels %s: %s", ground_truth.shape, ground_truth)
    LOG.info("Features %s: %s", features.shape, features.columns)

    gt_link = add_links_gt(features, ground_truth)

    if remove_no_overlap:
        features, gt_link = select_competing_links(features, gt_link)
        if features.empty:
            raise ValueError("Dataset is empty after removing no overlap links")

    classifier = _validate_classifier(classifier)

    classifier.fit(features, gt_link)

    if insert_prob:
        LOG.info(
            "Predicting links probabilities with features: %s", str(features.columns)
        )

        probs = classifier.predict_proba(features)

        LOG.info("Predicted probabilities %s: %s", probs.shape, probs)

        add_links_prob(
            config,
            np.asarray(features.index.to_list()),
            probs,
        )

    return classifier



import logging
from pathlib import Path
from typing import Optional, Union

import toml
from pydantic import BaseModel, ConfigDict, Field

from ultrack.config.dataconfig import DataConfig
from ultrack.config.segmentationconfig import SegmentationConfig
from ultrack.config.trackingconfig import TrackingConfig

LOG = logging.getLogger(__name__)


CFG_ALIAS_TO_ATTR = {
    "segmentation": "segmentation_config",
    "linking": "linking_config",
    "tracking": "tracking_config",
    "data": "data_config",
}


class LinkingConfig(BaseModel):
    """
    Candidate cell hypotheses linking configuration
    """

    max_distance: float = 15.0
    """Maximum distance between neighboring segments """

    n_workers: int = 1
    """Number of worker threads """

    max_neighbors: int = 5
    """Maximum number of neighbors per candidate segment """

    distance_weight: float = 0.0
    r"""
    Penalization weight :math:`\gamma` for distance between segment centroids,
    :math:`w_{pq} - \gamma \|c_p - c_q\|_2`, where :math:`c_p` is region :math:`p` center of mass
    """

    z_score_threshold: float = 5.0
    """
    ``SPECIAL``: z-score threshold between intensity values from within
    the segmentation masks of neighboring segments
    """

    model_config = ConfigDict(extra="forbid")


class MainConfig(BaseModel):
    data_config: Optional[DataConfig] = Field(
        default_factory=DataConfig,
        alias="data",
    )
    """
    Configuration for intermediate data storage and retrieval.
    """

    segmentation_config: SegmentationConfig = Field(
        default_factory=SegmentationConfig, alias="segmentation"
    )
    """Segmentation hypotheses creation configuration """

    linking_config: LinkingConfig = Field(
        default_factory=LinkingConfig, alias="linking"
    )
    """Candidate cell hypotheses linking configuration"""

    tracking_config: TrackingConfig = Field(
        default_factory=TrackingConfig, alias="tracking"
    )
    """Tracking (segmentation & linking selection) configuration """


def load_config(path: Union[str, Path]) -> MainConfig:
    """Creates MainConfig from TOML file."""
    with open(path) as f:
        data = toml.load(f)
        LOG.info(data)
        return MainConfig.model_validate(data)


def save_config(config, path: Union[str, Path]):
    """Saved MainConfig to TOML file."""
    with open(path, mode="w") as f:
        toml.dump(config.model_dump(by_alias=True), f)



import warnings
from pathlib import Path
from typing import Any, Callable, Optional, Tuple

import click
from toolz import curry

from ultrack.config import MainConfig, load_config


def paths_argument(**kwargs) -> Callable:
    def _decorator(f: Callable) -> Callable:
        return click.argument(
            "paths",
            nargs=-1,
            type=click.Path(path_type=Path),
            **kwargs,
        )(f)

    return _decorator


def _config_callback(ctx: click.Context, opt: click.Option, value: str) -> MainConfig:
    return load_config(value)


def config_option() -> Callable:
    def _decorator(f: Callable) -> Callable:
        return click.option(
            "--config",
            "-cfg",
            required=True,
            help="Ultrack configuration file (.toml)",
            callback=_config_callback,
        )(f)

    return _decorator


def _batch_index_callback(
    ctx: click.Context, opt: click.Option, value: Optional[int]
) -> Optional[int]:
    if value is not None and ctx.params.get("overwrite", False):
        warnings.warn(
            "ATTENTION: Overwriting while using batch indices will result in multiple deletions."
        )
    return value


def batch_index_option() -> Callable:
    def _decorator(f: Callable) -> Callable:
        return click.option(
            "--batch-index",
            "-b",
            required=False,
            default=None,
            show_default=True,
            type=int,
            callback=_batch_index_callback,
            help="Batch index to process a subset of time points. ATTENTION: this it not the time index.",
        )(f)

    return _decorator


def overwrite_option() -> Callable:
    def _decorator(f: Callable) -> Callable:
        return click.option(
            "--overwrite",
            "-ow",
            is_flag=True,
            default=False,
            type=bool,
            help="Enables overwriting existing data.",
            is_eager=True,
        )(f)

    return _decorator


def output_directory_option(**kwargs) -> Callable:

    if "required" not in kwargs and "default" not in kwargs:
        kwargs["required"] = True

    def _decorator(f: Callable) -> Callable:
        return click.option(
            "--output-directory",
            "-o",
            type=click.Path(path_type=Path),
            **kwargs,
        )(f)

    return _decorator


@curry
def tuple_callback(
    ctx: click.Context,
    opt: click.Option,
    value: str,
    dtype: Callable = int,
    length: Optional[int] = None,
) -> Optional[Tuple[Any]]:
    """Parses string to tuple given dtype and optional length.
       Returns None if None is supplied.
    Parameters
    ----------
    ctx : click.Context
        CLI context, not used.
    opt : click.Option
        CLI option, not used.
    value : str
        Input value.
    dtype : Callable, optional
        Data type for type casting, by default int
    length : Optional[int], optional
        Optional length for length checking, by default None
    Returns
    -------
    Tuple[Any]
        Tuple of given dtype and length (optional).
    """
    if value is None:
        return None
    tup = tuple(dtype(s) for s in value.split(","))
    if length is not None and length != len(tup):
        raise ValueError(f"Expected tuple of length {length}, got input {tup}")
    return tup


def napari_reader_option() -> Callable:
    def _decorator(f: Callable) -> Callable:
        return click.option(
            "--reader-plugin",
            "-r",
            default="napari",
            type=str,
            show_default=True,
            help="Napari reader plugin.",
        )(f)

    return _decorator


def layer_key_option() -> Callable:
    def _decorator(f: Callable) -> Callable:
        return click.option(
            "--layer-key",
            "-l",
            default=0,
            type=str,
            show_default=True,
            help="Layer key to index multi-channel input.",
        )(f)

    return _decorator


def channel_axis_option(default: Optional[int], help: str) -> Callable:
    def _decorator(f: Callable) -> Callable:
        return click.option(
            "--channel-axis",
            "-cha",
            required=False,
            type=int,
            default=default,
            show_default=True,
            help=help,
        )(f)

    return _decorator


def persistence_option() -> Callable:
    def _decorator(f: Callable) -> Callable:
        return click.option(
            "--persistence",
            is_flag=True,
            default=False,
            type=bool,
            help="Whether to include persistence features.",
        )(f)

    return _decorator



from contextlib import contextmanager

from qtpy.QtCore import Qt
from qtpy.QtGui import QCursor
from qtpy.QtWidgets import QApplication


@contextmanager
def wait_cursor():
    """Changes to wait cursor while inside context."""
    try:
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        yield
    finally:
        QApplication.restoreOverrideCursor()



from typing import List, Optional, Tuple, Union

import pandas as pd
from numpy.typing import ArrayLike
from scipy.optimize import linear_sum_assignment
from scipy.spatial.distance import cdist
from zarr.storage import StoreLike

from ultrack.utils.constants import NO_PARENT
from ultrack.utils.segmentation import SegmentationPainter, copy_segments


def tracks_starts(tracks_df: pd.DataFrame) -> pd.DataFrame:
    """
    Returns the starting nodes of each track in the given DataFrame.

    Parameters
    ----------
    tracks_df: pd.DataFrame
        The DataFrame containing track information.

    Returns
    -------
    pd.DataFrame
        The DataFrame containing the starting nodes of each track.

    """
    starting_tracklets = tracks_df[tracks_df["parent_track_id"] == NO_PARENT]
    starts = starting_tracklets.groupby("track_id").apply(
        lambda x: x.loc[x["t"].idxmin()]
    )
    return starts


def tracks_ends(tracks_df: pd.DataFrame) -> pd.DataFrame:
    """
    Find the ending nodes of tracks in the given DataFrame.

    Parameters
    ----------
    tracks_df : pd.DataFrame
        The DataFrame containing the tracks information.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the ending nodes of tracks.

    """
    ending_tracks = tracks_df[~tracks_df["track_id"].isin(tracks_df["parent_track_id"])]
    ends = ending_tracks.groupby("track_id").apply(lambda x: x.loc[x["t"].idxmax()])
    return ends


def update_track_id(
    tracks_df: pd.DataFrame,
    old_track_id: int,
    new_track_id: int,
    new_track_id_parent: int,
    segmentation_painter: Optional[SegmentationPainter] = None,
) -> None:
    """
    Update track_id in `tracks_df` from `old_track_id` to `new_track_id`.

    Parameters
    ----------
    tracks_df : pd.DataFrame
        DataFrame containing track information with columns:
            "track_id" : Unique identifier for each track.
            "parent_track_id" : Identifier of the parent track in the forest.
            (Other columns may be present in the DataFrame but are not used in this function.)
    old_track_id : int
        Old track id to be replaced.
    new_track_id : int
        New track id to replace the old one.
    new_track_id_parent : int
        Existing parent of new track id.
    segmentation_painter : Optional[SegmentationPainter]
        The painter to update the segments, if not provided the segments are not updated.
    """
    mask = tracks_df["track_id"] == old_track_id
    tracks_df.loc[mask, "track_id"] = new_track_id
    tracks_df.loc[mask, "parent_track_id"] = new_track_id_parent
    tracks_df.loc[tracks_df["parent_track_id"] == old_track_id, "parent_track_id"] = (
        new_track_id
    )

    if segmentation_painter is not None:
        for t in tracks_df.loc[mask, "t"]:
            segmentation_painter.add_relabel(
                time_pt=int(t), src_label=old_track_id, dst_label=new_track_id
            )


def _add_new_nodes(
    tracks_df: pd.DataFrame,
    new_nodes: List[pd.Series],
) -> pd.DataFrame:
    """
    Helper function to add `new_nodes` to `tracks_df`.
    """
    tracks_df = pd.concat([tracks_df, pd.DataFrame(new_nodes)], ignore_index=True)
    for int_cols in ("track_id", "parent_track_id"):
        tracks_df[int_cols] = tracks_df[int_cols].astype(int)

    tracks_df = tracks_df.sort_values(["track_id", "t"])

    return tracks_df


def close_tracks_gaps(
    tracks_df: pd.DataFrame,
    max_gap: int,
    max_radius: float,
    spatial_columns: List[str] = ["z", "y", "x"],
    scale: Optional[ArrayLike] = None,
    segments: Optional[ArrayLike] = None,
    segments_store_or_path: Optional[StoreLike] = None,
    overwrite: bool = False,
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, ArrayLike]]:
    """
    Close gaps between tracklets in the given DataFrame.

    Parameters
    ----------
    tracks_df : pd.DataFrame
        The DataFrame containing the tracks information.
    max_gap : int
        The maximum gap size to close.
    max_radius : float
        The maximum distance between the end of one tracklet and the start of the next tracklet.
    spatial_columns : List[str]
        The names of the columns containing the spatial information.
    scale : Optional[ArrayLike]
        The scaling factors for the spatial columns.
    segments : Optional[ArrayLike]
        When provided, the function will update the segments labels to match the tracks.
    segments_store_or_path : Union[Store, Path, str, None]
        The store or path to save the updated segments, if not provided in memory store is used.
    overwrite : bool
        If True, overwrites the segments store if it already exists.

    Returns
    -------
    Union[pd.DataFrame, Tuple[pd.DataFrame, ArrayLike]]
        The DataFrame containing the tracks information with the gaps closed.
    """

    if segments is None:
        out_segments = None
        segm_painter = None
    else:
        out_segments = copy_segments(segments, segments_store_or_path, overwrite)
        segm_painter = SegmentationPainter(out_segments)

    tracks_df = tracks_df.copy()

    if scale is not None:
        tracks_df[spatial_columns] *= scale

    starts = tracks_starts(tracks_df)
    ends = tracks_ends(tracks_df)

    new_nodes = []
    track_ids_in_queue = set()

    for gap in range(1, max_gap + 1):

        starts_by_t = starts.groupby("t")

        to_remove_from_start = []
        to_remove_from_end = []

        # backwards so multiple corrections are allowed
        for t, end_group in reversed(tuple(ends.groupby("t"))):

            try:
                start_group = starts_by_t.get_group(t + gap + 1)
            except KeyError:
                continue

            # optimal distance matching between end and start tracklets
            dists = cdist(end_group[spatial_columns], start_group[spatial_columns])
            row_ind, col_ind = linear_sum_assignment(dists)
            valid_matches = dists[row_ind, col_ind] < max_radius

            for i, j in zip(row_ind[valid_matches], col_ind[valid_matches]):
                end_node = end_group.iloc[i]
                start_node = start_group.iloc[j]

                if int(start_node["track_id"]) in track_ids_in_queue:
                    # forcing addition of new nodes to avoid inconsistencies
                    tracks_df = _add_new_nodes(tracks_df, new_nodes)
                    # reset
                    new_nodes = []
                    track_ids_in_queue = set()

                update_track_id(
                    tracks_df,
                    start_node["track_id"],
                    end_node["track_id"],
                    end_node["parent_track_id"],
                    segm_painter,
                )

                for t in range(int(end_node["t"] + 1), int(start_node["t"])):
                    new_node = end_node.copy()
                    weight = (t - end_node["t"]) / (start_node["t"] - end_node["t"])
                    step = (
                        start_node[spatial_columns] - end_node[spatial_columns]
                    ) * weight
                    new_node["t"] = t
                    new_node[spatial_columns] = end_node[spatial_columns] + step
                    new_nodes.append(new_node)
                    track_ids_in_queue.add(int(new_node["track_id"].item()))

                    if segm_painter is not None:
                        segm_painter.add_new(
                            src_time_pt=int(end_node["t"]),
                            src_label=int(end_node["track_id"]),
                            dst_time_pt=t,
                            dst_label=int(new_node["track_id"]),
                            shift=step,
                        )

            to_remove_from_start.extend(start_group.index[col_ind[valid_matches]])
            to_remove_from_end.extend(end_group.index[row_ind[valid_matches]])

        starts = starts.drop(to_remove_from_start)
        ends = ends.drop(to_remove_from_end)

    # adding remaining nodes
    tracks_df = _add_new_nodes(tracks_df, new_nodes)

    if scale is not None:
        tracks_df[spatial_columns] /= scale

    if out_segments is None:
        return tracks_df

    segm_painter.apply_changes()

    return tracks_df, out_segments



from magicgui.widgets import FileEdit

from ultrack.widgets.ultrackwidget._legacy.baseconfigwidget import BaseConfigWidget


class DataWidget(BaseConfigWidget):
    def _setup_widgets(self) -> None:
        self._attr_to_widget = {
            "working_dir": FileEdit(mode="d", label="Working dir."),
        }

        for widget in self._attr_to_widget.values():
            self.append(widget)
